# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from collections import namedtuple
from logging import DEBUG as _DEBUG
from logging import getLogger

from . import hidpp10 as _hidpp10
from . import hidpp20 as _hidpp20
from . import special_keys as _special_keys
from .common import NamedInt as _NamedInt
from .common import NamedInts as _NamedInts
from .common import bytes2int as _bytes2int
from .common import int2bytes as _int2bytes
from .common import unpack as _unpack
from .i18n import _
from .settings import KIND as _KIND
from .settings import BitFieldSetting as _BitFieldSetting
from .settings import BitFieldValidator as _BitFieldV
from .settings import BooleanValidator as _BooleanV
from .settings import ChoicesMapValidator as _ChoicesMapV
from .settings import ChoicesValidator as _ChoicesV
from .settings import FeatureRW as _FeatureRW
from .settings import FeatureRWMap as _FeatureRWMap
from .settings import RangeValidator as _RangeV
from .settings import RegisterRW as _RegisterRW
from .settings import Setting as _Setting
from .settings import Settings as _Settings

_log = getLogger(__name__)
del getLogger

_DK = _hidpp10.DEVICE_KIND
_R = _hidpp10.REGISTERS
_F = _hidpp20.FEATURE

#
# pre-defined basic setting descriptors
#


def register_toggle(
    name,
    register,
    true_value=_BooleanV.default_true,
    false_value=_BooleanV.default_false,
    mask=_BooleanV.default_mask,
    label=None,
    description=None,
    device_kind=None
):
    validator = _BooleanV(true_value=true_value, false_value=false_value, mask=mask)
    rw = _RegisterRW(register)
    return _Setting(name, rw, validator, label=label, description=description, device_kind=device_kind)


def register_choices(name, register, choices, kind=_KIND.choice, label=None, description=None, device_kind=None):
    assert choices
    validator = _ChoicesV(choices)
    rw = _RegisterRW(register)
    return _Setting(name, rw, validator, kind=kind, label=label, description=description, device_kind=device_kind)


def feature_toggle(
    name,
    feature,
    read_fnid=_FeatureRW.default_read_fnid,
    write_fnid=_FeatureRW.default_write_fnid,
    true_value=_BooleanV.default_true,
    false_value=_BooleanV.default_false,
    mask=_BooleanV.default_mask,
    label=None,
    description=None,
    device_kind=None
):
    validator = _BooleanV(true_value=true_value, false_value=false_value, mask=mask)
    rw = _FeatureRW(feature, read_fnid=read_fnid, write_fnid=write_fnid)
    return _Setting(name, rw, validator, feature=feature, label=label, description=description, device_kind=device_kind)


def feature_bitfield_toggle(
    name,
    feature,
    options,
    read_fnid=_FeatureRW.default_read_fnid,
    write_fnid=_FeatureRW.default_write_fnid,
    label=None,
    description=None,
    device_kind=None
):
    assert options
    validator = _BitFieldV(options)
    rw = _FeatureRW(feature, read_fnid=read_fnid, write_fnid=write_fnid)
    return _BitFieldSetting(
        name, rw, validator, feature=feature, label=label, description=description, device_kind=device_kind
    )


def feature_bitfield_toggle_dynamic(
    name,
    feature,
    options_callback,
    read_fnid=_FeatureRW.default_read_fnid,
    write_fnid=_FeatureRW.default_write_fnid,
    label=None,
    description=None,
    device_kind=None
):
    def instantiate(device):
        options = options_callback(device)
        setting = feature_bitfield_toggle(
            name,
            feature,
            options,
            read_fnid=read_fnid,
            write_fnid=write_fnid,
            label=label,
            description=description,
            device_kind=device_kind
        )
        return setting(device)

    instantiate._rw_kind = _FeatureRW.kind
    return instantiate


def feature_choices(name, feature, choices, **kwargs):
    assert choices
    validator = _ChoicesV(choices, **kwargs)
    rw = _FeatureRW(feature, **kwargs)
    return _Setting(name, rw, validator, feature=feature, kind=_KIND.choice, **kwargs)


def feature_choices_dynamic(name, feature, choices_callback, **kwargs):
    # Proxy that obtains choices dynamically from a device
    def instantiate(device):
        # Obtain choices for this feature
        choices = choices_callback(device)
        if not choices:  # no choices, so don't create a setting
            return None
        setting = feature_choices(name, feature, choices, **kwargs)
        return setting(device)

    instantiate._rw_kind = _FeatureRW.kind
    return instantiate


# maintain a mapping from keys (NamedInts) to one of a list of choices (NamedInts), default is first one
# the setting is stored as a JSON-compatible object mapping the key int (as a string) to the choice int
def feature_map_choices(name, feature, choicesmap, **kwargs):
    assert choicesmap
    validator = _ChoicesMapV(choicesmap, **kwargs)
    rw = _FeatureRWMap(feature, **kwargs)
    return _Settings(name, rw, validator, feature=feature, kind=_KIND.map_choice, **kwargs)


def feature_map_choices_dynamic(name, feature, choices_callback, **kwargs):
    # Proxy that obtains choices dynamically from a device
    def instantiate(device):
        choices = choices_callback(device)
        if not choices:  # no choices, so don't create a Setting
            return None
        setting = feature_map_choices(name, feature, choices, **kwargs)
        return setting(device)

    instantiate._rw_kind = _FeatureRWMap.kind
    return instantiate


def feature_range(
    name,
    feature,
    min_value,
    max_value,
    read_fnid=_FeatureRW.default_read_fnid,
    write_fnid=_FeatureRW.default_write_fnid,
    rw=None,
    bytes_count=None,
    label=None,
    description=None,
    device_kind=None
):
    validator = _RangeV(min_value, max_value, bytes_count=bytes_count)
    if rw is None:
        rw = _FeatureRW(feature, read_fnid=read_fnid, write_fnid=write_fnid)
    return _Setting(
        name, rw, validator, feature=feature, kind=_KIND.range, label=label, description=description, device_kind=device_kind
    )


#
# common strings for settings - name, string to display in main window, tool tip for main window
#

_HAND_DETECTION = ('hand-detection', _('Hand Detection'), _('Turn on illumination when the hands hover over the keyboard.'))
_SMOOTH_SCROLL = ('smooth-scroll', _('Smooth Scrolling'), _('High-sensitivity mode for vertical scroll with the wheel.'))
_SIDE_SCROLL = (
    'side-scroll', _('Side Scrolling'),
    _(
        'When disabled, pushing the wheel sideways sends custom button events\n'
        'instead of the standard side-scrolling events.'
    )
)
_HI_RES_SCROLL = (
    'hi-res-scroll', _('High Resolution Scrolling'), _('High-sensitivity mode for vertical scroll with the wheel.')
)
_LOW_RES_SCROLL = (
    'lowres-smooth-scroll', _('HID++ Scrolling'),
    _('HID++ mode for vertical scroll with the wheel.') + '\n' + _('Effectively turns off wheel scrolling in Linux.')
)
_HIRES_INV = (
    'hires-smooth-invert', _('High Resolution Wheel Invert'),
    _('High-sensitivity wheel invert direction for vertical scroll.')
)
_HIRES_RES = ('hires-smooth-resolution', _('Wheel Resolution'), _('High-sensitivity mode for vertical scroll with the wheel.'))
_FN_SWAP = (
    'fn-swap', _('Swap Fx function'),
    _(
        'When set, the F1..F12 keys will activate their special function,\n'
        'and you must hold the FN key to activate their standard function.'
    ) + '\n\n' + _(
        'When unset, the F1..F12 keys will activate their standard function,\n'
        'and you must hold the FN key to activate their special function.'
    )
)
_DPI = ('dpi', _('Sensitivity (DPI)'), None)
_POINTER_SPEED = (
    'pointer_speed', _('Sensitivity (Pointer Speed)'), _('Speed multiplier for mouse (256 is normal multiplier).')
)
_SMART_SHIFT = (
    'smart-shift', _('Smart Shift'),
    _(
        'Automatically switch the mouse wheel between ratchet and freespin mode.\n'
        'The mouse wheel is always free at 0, and always locked at 50'
    )
)
_BACKLIGHT = ('backlight', _('Backlight'), _('Turn illumination on or off on keyboard.'))
_REPROGRAMMABLE_KEYS = (
    'reprogrammable-keys', _('Actions'), _('Change the action for the key or button.') + '\n' +
    _('Changing important actions (such as for the left mouse button) can result in an unusable system.')
)
_DISABLE_KEYS = ('disable-keyboard-keys', _('Disable keys'), _('Disable specific keyboard keys.'))
_PLATFORM = ('multiplatform', _('Set OS'), _('Change keys to match OS.'))
_CHANGE_HOST = ('change-host', _('Change Host'), _('Switch connection to a different host'))
_THUMB_SCROLL_MODE = (
    'thumb-scroll-mode', _('HID++ Thumb Scrolling'),
    _('HID++ mode for horizontal scroll with the thumb wheel.') + '\n' + _('Effectively turns off thumb scrolling in Linux.')
)
_THUMB_SCROLL_INVERT = ('thumb-scroll-invert', _('Thumb Scroll Invert'), _('Invert thumb scroll direction.'))

#
# Keyword arguments for setting template functions:
#  label='', description='' - label and tooltip to be shown in GUI
#  persist=True - whether to store the values and reapply them from now on
#  device_kind - the kinds of devices that setting is suitable for (NOT CURRENTLY USED)
#  read_fnid=0x00, write_fnid=0x10 - default 0x00 and 0x10 function numbers (times 16) to read and write setting
#  bytes_count=1 - number of bytes for the data (ignoring the key, if any)
# only for boolean settings
#  true_value=0x01, false_value=0x00,  mask=0xFF - integer or byte strings for boolean settings
# only for map choices
#  key_bytes_count=1 - number of bytes in the key
#  extra_default - extra value that cannot be set but means same as default value
# only for choices and map choices
#  read_skip_bytes_count=0 - number of bytes to ignore before the data when reading
#  write_prefix_bytes=b'' - bytes to put before the data writing


def _register_hand_detection(
    register=_R.keyboard_hand_detection, true_value=b'\x00\x00\x00', false_value=b'\x00\x00\x30', mask=b'\x00\x00\xFF'
):
    return register_toggle(
        _HAND_DETECTION[0],
        register,
        true_value=true_value,
        false_value=false_value,
        label=_HAND_DETECTION[1],
        description=_HAND_DETECTION[2],
        device_kind=(_DK.keyboard, )
    )


def _register_fn_swap(register=_R.keyboard_fn_swap, true_value=b'\x00\x01', mask=b'\x00\x01'):
    return register_toggle(
        _FN_SWAP[0],
        register,
        true_value=true_value,
        mask=mask,
        label=_FN_SWAP[1],
        description=_FN_SWAP[2],
        device_kind=(_DK.keyboard, )
    )


def _register_smooth_scroll(register=_R.mouse_button_flags, true_value=0x40, mask=0x40):
    return register_toggle(
        _SMOOTH_SCROLL[0],
        register,
        true_value=true_value,
        mask=mask,
        label=_SMOOTH_SCROLL[1],
        description=_SMOOTH_SCROLL[2],
        device_kind=(_DK.mouse, _DK.trackball)
    )


def _register_side_scroll(register=_R.mouse_button_flags, true_value=0x02, mask=0x02):
    return register_toggle(
        _SIDE_SCROLL[0],
        register,
        true_value=true_value,
        mask=mask,
        label=_SIDE_SCROLL[1],
        description=_SIDE_SCROLL[2],
        device_kind=(_DK.mouse, _DK.trackball)
    )


def _register_dpi(register=_R.mouse_dpi, choices=None):
    return register_choices(
        _DPI[0], register, choices, label=_DPI[1], description=_DPI[2], device_kind=(_DK.mouse, _DK.trackball)
    )


def _feature_fn_swap():
    return feature_toggle(
        _FN_SWAP[0], _F.FN_INVERSION, label=_FN_SWAP[1], description=_FN_SWAP[2], device_kind=(_DK.keyboard, )
    )


# this might not be correct for this feature
def _feature_new_fn_swap():
    return feature_toggle(
        _FN_SWAP[0], _F.NEW_FN_INVERSION, label=_FN_SWAP[1], description=_FN_SWAP[2], device_kind=(_DK.keyboard, )
    )


# ignore the capabilities part of the feature - all devices should be able to swap Fn state
# just use the current host (first byte = 0xFF) part of the feature to read and set the Fn state
def _feature_k375s_fn_swap():
    return feature_toggle(
        _FN_SWAP[0],
        _F.K375S_FN_INVERSION,
        label=_FN_SWAP[1],
        description=_FN_SWAP[2],
        true_value=b'\xFF\x01',
        false_value=b'\xFF\x00',
        device_kind=(_DK.keyboard, )
    )


# FIXME: This will enable all supported backlight settings,
# we should allow the users to select which settings they want to enable.
def _feature_backlight2():
    return feature_toggle(
        _BACKLIGHT[0], _F.BACKLIGHT2, label=_BACKLIGHT[1], description=_BACKLIGHT[2], device_kind=(_DK.keyboard, )
    )


def _feature_hi_res_scroll():
    return feature_toggle(
        _HI_RES_SCROLL[0],
        _F.HI_RES_SCROLLING,
        label=_HI_RES_SCROLL[1],
        description=_HI_RES_SCROLL[2],
        device_kind=(_DK.mouse, _DK.trackball)
    )


def _feature_lowres_smooth_scroll():
    return feature_toggle(
        _LOW_RES_SCROLL[0],
        _F.LOWRES_WHEEL,
        label=_LOW_RES_SCROLL[1],
        description=_LOW_RES_SCROLL[2],
        device_kind=(_DK.mouse, _DK.trackball)
    )


def _feature_hires_smooth_invert():
    return feature_toggle(
        _HIRES_INV[0],
        _F.HIRES_WHEEL,
        read_fnid=0x10,
        write_fnid=0x20,
        true_value=0x04,
        mask=0x04,
        label=_HIRES_INV[1],
        description=_HIRES_INV[2],
        device_kind=(_DK.mouse, _DK.trackball)
    )


def _feature_hires_smooth_resolution():
    return feature_toggle(
        _HIRES_RES[0],
        _F.HIRES_WHEEL,
        read_fnid=0x10,
        write_fnid=0x20,
        true_value=0x02,
        mask=0x02,
        label=_HIRES_RES[1],
        description=_HIRES_RES[2],
        device_kind=(_DK.mouse, _DK.trackball)
    )


def _feature_smart_shift():
    _MIN_SMART_SHIFT_VALUE = 0
    _MAX_SMART_SHIFT_VALUE = 50

    class _SmartShiftRW(_FeatureRW):
        def __init__(self, feature):
            super(_SmartShiftRW, self).__init__(feature)

        def read(self, device):
            value = super(_SmartShiftRW, self).read(device)
            if _bytes2int(value[0:1]) == 1:
                # Mode = Freespin, map to minimum
                return _int2bytes(_MIN_SMART_SHIFT_VALUE, count=1)
            else:
                # Mode = smart shift, map to the value, capped at maximum
                threshold = min(_bytes2int(value[1:2]), _MAX_SMART_SHIFT_VALUE)
                return _int2bytes(threshold, count=1)

        def write(self, device, data_bytes):
            threshold = _bytes2int(data_bytes)
            # Freespin at minimum
            mode = 1 if threshold == _MIN_SMART_SHIFT_VALUE else 2

            # Ratchet at maximum
            if threshold == _MAX_SMART_SHIFT_VALUE:
                threshold = 255

            data = _int2bytes(mode, count=1) + _int2bytes(threshold, count=1) * 2
            return super(_SmartShiftRW, self).write(device, data)

    return feature_range(
        _SMART_SHIFT[0],
        _F.SMART_SHIFT,
        _MIN_SMART_SHIFT_VALUE,
        _MAX_SMART_SHIFT_VALUE,
        bytes_count=1,
        rw=_SmartShiftRW(_F.SMART_SHIFT),
        label=_SMART_SHIFT[1],
        description=_SMART_SHIFT[2],
        device_kind=(_DK.mouse, _DK.trackball)
    )


def _feature_adjustable_dpi_choices(device):
    # [1] getSensorDpiList(sensorIdx)
    reply = device.feature_request(_F.ADJUSTABLE_DPI, 0x10)
    # Should not happen, but might happen when the user unplugs device while the
    # query is being executed. TODO retry logic?
    assert reply, 'Oops, DPI list cannot be retrieved!'
    dpi_list = []
    step = None
    for val in _unpack('!7H', reply[1:1 + 14]):
        if val == 0:
            break
        if val >> 13 == 0b111:
            assert step is None and len(dpi_list) == 1, \
              'Invalid DPI list item: %r' % val
            step = val & 0x1fff
        else:
            dpi_list.append(val)
    if step:
        assert len(dpi_list) == 2, 'Invalid DPI list range: %r' % dpi_list
        dpi_list = range(dpi_list[0], dpi_list[1] + 1, step)
    return _NamedInts.list(dpi_list)


def _feature_adjustable_dpi():
    """Pointer Speed feature"""
    # Assume sensorIdx 0 (there is only one sensor)
    # [2] getSensorDpi(sensorIdx) -> sensorIdx, dpiMSB, dpiLSB
    # [3] setSensorDpi(sensorIdx, dpi)
    return feature_choices_dynamic(
        _DPI[0],
        _F.ADJUSTABLE_DPI,
        _feature_adjustable_dpi_choices,
        read_fnid=0x20,
        write_fnid=0x30,
        bytes_count=3,
        label=_DPI[1],
        description=_DPI[2],
        device_kind=(_DK.mouse, _DK.trackball)
    )


def _feature_pointer_speed():
    """Pointer Speed feature"""
    # min and max values taken from usb traces of Win software
    return feature_range(
        _POINTER_SPEED[0],
        _F.POINTER_SPEED,
        0x002e,
        0x01ff,
        read_fnid=0x0,
        write_fnid=0x10,
        bytes_count=2,
        label=_POINTER_SPEED[1],
        description=_POINTER_SPEED[2],
        device_kind=(_DK.mouse, _DK.trackball)
    )


# the keys for the choice map are Logitech controls (from special_keys)
# each choice value is a NamedInt with the string from a task (to be shown to the user)
# and the integer being the control number for that task (to be written to the device)
# Solaar only remaps keys (controlled by key gmask and group), not other key reprogramming
def _feature_reprogrammable_keys_choices(device):
    count = device.feature_request(_F.REPROG_CONTROLS_V4)
    assert count, 'Oops, reprogrammable key count cannot be retrieved!'
    count = ord(count[:1])  # the number of key records
    keys = [None] * count
    groups = [[] for i in range(0, 9)]
    choices = {}
    for i in range(0, count):  # get the data for each key record on device
        keydata = device.feature_request(_F.REPROG_CONTROLS_V4, 0x10, i)
        key, key_task, flags, pos, group, gmask = _unpack('!HHBBBB', keydata[:8])
        action = _NamedInt(key, str(_special_keys.TASK[key_task]))
        keys[i] = (_special_keys.CONTROL[key], action, flags, gmask)
        groups[group].append(action)
    for k in keys:
        # if k[2] & _special_keys.KEY_FLAG.reprogrammable:  # this flag is only to show in UI, ignore in Solaar
        if k[3]:  # only keys with a non-zero gmask are remappable
            key_choices = [k[1]]  # it should always be possible to map the key to itself
            for g in range(1, 9):  # group 0 and gmask 0 (k[3]) does not indicate remappability so don't consider group 0
                if (k[3] == 0 if g == 0 else k[3] & 2**(g - 1)):
                    for gm in groups[g]:
                        if int(gm) != int(k[0]):  # don't put itself in twice
                            key_choices.append(gm)
            if len(key_choices) > 1:
                choices[k[0]] = key_choices
    return choices


def _feature_reprogrammable_keys():
    return feature_map_choices_dynamic(
        _REPROGRAMMABLE_KEYS[0],
        _F.REPROG_CONTROLS_V4,
        _feature_reprogrammable_keys_choices,
        read_fnid=0x20,
        write_fnid=0x30,
        key_bytes_count=2,
        bytes_count=2,
        read_skip_bytes_count=1,
        write_prefix_bytes=b'\x00',
        extra_default=0,
        label=_REPROGRAMMABLE_KEYS[1],
        description=_REPROGRAMMABLE_KEYS[2],
        device_kind=(_DK.keyboard, ),
    )


def _feature_disable_keyboard_keys_key_list(device):
    mask = device.feature_request(_F.KEYBOARD_DISABLE_KEYS)[0]
    options = [_special_keys.DISABLE[1 << i] for i in range(8) if mask & (1 << i)]
    return options


def _feature_disable_keyboard_keys():
    return feature_bitfield_toggle_dynamic(
        _DISABLE_KEYS[0],
        _F.KEYBOARD_DISABLE_KEYS,
        _feature_disable_keyboard_keys_key_list,
        read_fnid=0x10,
        write_fnid=0x20,
        label=_DISABLE_KEYS[1],
        description=_DISABLE_KEYS[2],
        device_kind=(_DK.keyboard, )
    )


# muultiplatform OS bits
OSS = [('Linux', 0x0400), ('MacOS', 0x2000), ('Windows', 0x0100), ('iOS', 0x4000), ('Android', 0x1000), ('WebOS', 0x8000),
       ('Chrome', 0x0800), ('WinEmb', 0x0200), ('Tizen', 0x0001)]


def _feature_multiplatform_choices(device):
    def _str_os_versions(low, high):
        def _str_os_version(version):
            if version == 0:
                return ''
            elif version & 0xFF:
                return str(version >> 8) + '.' + str(version & 0xFF)
            else:
                return str(version >> 8)

        return '' if low == 0 and high == 0 else ' ' + _str_os_version(low) + '-' + _str_os_version(high)

    infos = device.feature_request(_F.MULTIPLATFORM)
    assert infos, 'Oops, multiplatform count cannot be retrieved!'
    flags, _ignore, num_descriptors = _unpack('!BBB', infos[:3])
    if not (flags & 0x02):  # can't set platform so don't create setting
        return []
    descriptors = []
    for index in range(0, num_descriptors):
        descriptor = device.feature_request(_F.MULTIPLATFORM, 0x10, index)
        platform, _ignore, os_flags, low, high = _unpack('!BBHHH', descriptor[:8])
        descriptors.append((platform, os_flags, low, high))
    choices = _NamedInts()
    for os_name, os_bit in OSS:
        for platform, os_flags, low, high in descriptors:
            os = os_name + _str_os_versions(low, high)
            if os_bit & os_flags and platform not in choices and os not in choices:
                choices[platform] = os
    return choices


def _feature_multiplatform():
    return feature_choices_dynamic(
        _PLATFORM[0],
        _F.MULTIPLATFORM,
        _feature_multiplatform_choices,
        read_fnid=0x00,
        read_skip_bytes_count=6,
        write_fnid=0x30,
        write_prefix_bytes=b'\xff',
        label=_PLATFORM[1],
        description=_PLATFORM[2]
    )


PLATFORMS = _NamedInts()
PLATFORMS[0x00] = 'iOS, MacOS'
PLATFORMS[0x01] = 'Android, Windows'


def _feature_dualplatform():
    return feature_choices(
        _PLATFORM[0],
        _F.DUALPLATFORM,
        PLATFORMS,
        read_fnid=0x10,
        write_fnid=0x20,
        label=_PLATFORM[1],
        description=_PLATFORM[2],
        device_kind=(_DK.keyboard, )
    )


def _feature_change_host_choices(device):
    infos = device.feature_request(_F.CHANGE_HOST)
    assert infos, 'Oops, host count cannot be retrieved!'
    numHosts, currentHost = _unpack('!BB', infos[:2])
    hostNames = _hidpp20.get_host_names(device)
    hostNames = hostNames if hostNames is not None else {}
    if currentHost not in hostNames or hostNames[currentHost][1] == '':
        import socket  # find name of current host and use it
        hostNames[currentHost] = (True, socket.gethostname().partition('.')[0])
    choices = _NamedInts()
    for host in range(0, numHosts):
        _ignore, hostName = hostNames.get(host, (False, ''))
        choices[host] = str(host + 1) + ':' + hostName if hostName else str(host + 1)
    return choices


def _feature_change_host():
    return feature_choices_dynamic(
        _CHANGE_HOST[0],
        _F.CHANGE_HOST,
        _feature_change_host_choices,
        persist=False,
        no_reply=True,
        read_fnid=0x00,
        read_skip_bytes_count=1,
        write_fnid=0x10,
        label=_CHANGE_HOST[1],
        description=_CHANGE_HOST[2]
    )


def _feature_thumb_mode():
    return feature_toggle(
        _THUMB_SCROLL_MODE[0],
        _F.THUMB_WHEEL,
        read_fnid=0x10,
        write_fnid=0x20,
        true_value=b'\x01\x00',
        false_value=b'\x00\x00',
        mask=b'\x01\x00',
        label=_THUMB_SCROLL_MODE[1],
        description=_THUMB_SCROLL_MODE[2],
        device_kind=(_DK.mouse, _DK.trackball)
    )


def _feature_thumb_invert():
    return feature_toggle(
        _THUMB_SCROLL_INVERT[0],
        _F.THUMB_WHEEL,
        read_fnid=0x10,
        write_fnid=0x20,
        true_value=b'\x00\x01',
        false_value=b'\x00\x00',
        mask=b'\x00\x01',
        label=_THUMB_SCROLL_INVERT[1],
        description=_THUMB_SCROLL_INVERT[2],
        device_kind=(_DK.mouse, _DK.trackball)
    )


#
#
#


def _S(name, featureID=None, featureFn=None, registerFn=None, identifier=None):
    return (name, featureID, featureFn, registerFn, identifier if identifier else name.replace('-', '_'))


_SETTINGS_TABLE = [
    _S(_HAND_DETECTION[0], registerFn=_register_hand_detection),
    _S(_SMOOTH_SCROLL[0], registerFn=_register_smooth_scroll),
    _S(_SIDE_SCROLL[0], registerFn=_register_side_scroll),
    _S(_HI_RES_SCROLL[0], _F.HI_RES_SCROLLING, _feature_hi_res_scroll),
    _S(_LOW_RES_SCROLL[0], _F.LOWRES_WHEEL, _feature_lowres_smooth_scroll),
    _S(_HIRES_INV[0], _F.HIRES_WHEEL, _feature_hires_smooth_invert),
    _S(_HIRES_RES[0], _F.HIRES_WHEEL, _feature_hires_smooth_resolution),
    _S(_FN_SWAP[0], _F.FN_INVERSION, _feature_fn_swap, registerFn=_register_fn_swap),
    _S(_FN_SWAP[0], _F.NEW_FN_INVERSION, _feature_new_fn_swap, identifier='new_fn_swap'),
    _S(_FN_SWAP[0], _F.K375S_FN_INVERSION, _feature_k375s_fn_swap, identifier='k375s_fn_swap'),
    _S(_DPI[0], _F.ADJUSTABLE_DPI, _feature_adjustable_dpi, registerFn=_register_dpi),
    _S(_POINTER_SPEED[0], _F.POINTER_SPEED, _feature_pointer_speed),
    _S(_SMART_SHIFT[0], _F.SMART_SHIFT, _feature_smart_shift),
    _S(_BACKLIGHT[0], _F.BACKLIGHT2, _feature_backlight2),
    _S(_REPROGRAMMABLE_KEYS[0], _F.REPROG_CONTROLS_V4, _feature_reprogrammable_keys),
    _S(_DISABLE_KEYS[0], _F.KEYBOARD_DISABLE_KEYS, _feature_disable_keyboard_keys),
    _S(_PLATFORM[0], _F.MULTIPLATFORM, _feature_multiplatform),
    _S(_PLATFORM[0], _F.DUALPLATFORM, _feature_dualplatform, identifier='dualplatform'),
    _S(_CHANGE_HOST[0], _F.CHANGE_HOST, _feature_change_host),
    _S(_THUMB_SCROLL_MODE[0], _F.THUMB_WHEEL, _feature_thumb_mode),
    _S(_THUMB_SCROLL_INVERT[0], _F.THUMB_WHEEL, _feature_thumb_invert),
]

_SETTINGS_LIST = namedtuple('_SETTINGS_LIST', [s[4] for s in _SETTINGS_TABLE])
RegisterSettings = _SETTINGS_LIST._make([s[3] for s in _SETTINGS_TABLE])
FeatureSettings = _SETTINGS_LIST._make([s[2] for s in _SETTINGS_TABLE])

del _SETTINGS_LIST

#
#
#


def check_feature(device, name, featureId, featureFn):
    """
    :param name: name for the setting
    :param featureId: the numeric Feature ID for this setting implementation
    :param featureFn: the function for this setting implementation
    """
    if featureId not in device.features:
        return
    try:
        detected = featureFn()(device)
        if _log.isEnabledFor(_DEBUG):
            _log.debug('check_feature[%s] detected %s', featureId, detected)
        return detected
    except Exception:
        from traceback import format_exc
        _log.error('check_feature[%s] inconsistent feature %s', featureId, format_exc())


# Returns True if device was queried to find features, False otherwise
def check_feature_settings(device, already_known):
    """Try to auto-detect device settings by the HID++ 2.0 features they have."""
    if device.features is None or not device.online:
        return False
    if device.protocol and device.protocol < 2.0:
        return False
    for name, featureId, featureFn, __, __ in _SETTINGS_TABLE:
        if featureId and featureFn:
            if not any(s.name == name for s in already_known):
                setting = check_feature(device, name, featureId, featureFn)
                if setting:
                    already_known.append(setting)
    return True


def check_feature_setting(device, setting_name):
    for name, featureId, featureFn, __, __ in _SETTINGS_TABLE:
        if name == setting_name and featureId and featureFn:
            return check_feature(device, name, featureId, featureFn)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import sys
import time

from binascii import hexlify, unhexlify
from select import select as _select
from threading import Lock

import hidapi as _hid

#
#
#

try:
    read_packet = raw_input
except NameError:
    # Python 3 equivalent of raw_input
    read_packet = input

interactive = os.isatty(0)
prompt = '?? Input: ' if interactive else ''
start_time = time.time()

strhex = lambda d: hexlify(d).decode('ascii').upper()
try:
    unicode  # noqa: F821
    # this is certanly Python 2
    is_string = lambda d: isinstance(d, unicode)  # noqa: F821
    # no easy way to distinguish between b'' and '' :(
    # or (isinstance(d, str) \
    #     and not any((chr(k) in d for k in range(0x00, 0x1F))) \
    #     and not any((chr(k) in d for k in range(0x80, 0xFF))) \
    #     )
except Exception:
    # this is certanly Python 3
    # In Py3, unicode and str are equal (the unicode object does not exist)
    is_string = lambda d: isinstance(d, str)

#
#
#

print_lock = Lock()
del Lock


def _print(marker, data, scroll=False):
    t = time.time() - start_time
    if is_string(data):
        s = marker + ' ' + data
    else:
        hexs = strhex(data)
        s = '%s (% 8.3f) [%s %s %s %s] %s' % (marker, t, hexs[0:2], hexs[2:4], hexs[4:8], hexs[8:], repr(data))

    with print_lock:
        # allow only one thread at a time to write to the console, otherwise
        # the output gets garbled, especially with ANSI codes.

        if interactive and scroll:
            # scroll the entire screen above the current line up by 1 line
            sys.stdout.write(
                '\033[s'  # save cursor position
                '\033[S'  # scroll up
                '\033[A'  # cursor up
                '\033[L'  # insert 1 line
                '\033[G'
            )  # move cursor to column 1
        sys.stdout.write(s)
        if interactive and scroll:
            # restore cursor position
            sys.stdout.write('\033[u')
        else:
            sys.stdout.write('\n')

        # flush stdout manually...
        # because trying to open stdin/out unbuffered programmatically
        # works much too differently in Python 2/3
        sys.stdout.flush()


def _error(text, scroll=False):
    _print('!!', text, scroll)


def _continuous_read(handle, timeout=2000):
    while True:
        try:
            reply = _hid.read(handle, 128, timeout)
        except OSError as e:
            _error('Read failed, aborting: ' + str(e), True)
            break
        assert reply is not None
        if reply:
            _print('>>', reply, True)


def _validate_input(line, hidpp=False):
    try:
        data = unhexlify(line.encode('ascii'))
    except Exception as e:
        _error('Invalid input: ' + str(e))
        return None

    if hidpp:
        if len(data) < 4:
            _error('Invalid HID++ request: need at least 4 bytes')
            return None
        if data[:1] not in b'\x10\x11':
            _error('Invalid HID++ request: first byte must be 0x10 or 0x11')
            return None
        if data[1:2] not in b'\xFF\x01\x02\x03\x04\x05\x06':
            _error('Invalid HID++ request: second byte must be 0xFF or one of 0x01..0x06')
            return None
        if data[:1] == b'\x10':
            if len(data) > 7:
                _error('Invalid HID++ request: maximum length of a 0x10 request is 7 bytes')
                return None
            while len(data) < 7:
                data = (data + b'\x00' * 7)[:7]
        elif data[:1] == b'\x11':
            if len(data) > 20:
                _error('Invalid HID++ request: maximum length of a 0x11 request is 20 bytes')
                return None
            while len(data) < 20:
                data = (data + b'\x00' * 20)[:20]

    return data


def _open(args):
    device = args.device
    if args.hidpp and not device:
        for d in _hid.enumerate(vendor_id=0x046d):
            if d.driver == 'logitech-djreceiver':
                device = d.path
                break
        if not device:
            sys.exit('!! No HID++ receiver found.')
    if not device:
        sys.exit('!! Device path required.')

    print('.. Opening device', device)
    handle = _hid.open_path(device)
    if not handle:
        sys.exit('!! Failed to open %s, aborting.' % device)

    print(
        '.. Opened handle %r, vendor %r product %r serial %r.' %
        (handle, _hid.get_manufacturer(handle), _hid.get_product(handle), _hid.get_serial(handle))
    )
    if args.hidpp:
        if _hid.get_manufacturer(handle) != b'Logitech':
            sys.exit('!! Only Logitech devices support the HID++ protocol.')
        print('.. HID++ validation enabled.')
    else:
        if (_hid.get_manufacturer(handle) == b'Logitech' and b'Receiver' in _hid.get_product(handle)):
            args.hidpp = True
            print('.. Logitech receiver detected, HID++ validation enabled.')

    return handle


#
#
#


def _parse_arguments():
    import argparse
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('--history', help='history file (default ~/.hidconsole-history)')
    arg_parser.add_argument('--hidpp', action='store_true', help='ensure input data is a valid HID++ request')
    arg_parser.add_argument(
        'device',
        nargs='?',
        help='linux device to connect to (/dev/hidrawX); '
        'may be omitted if --hidpp is given, in which case it looks for the first Logitech receiver'
    )
    return arg_parser.parse_args()


def main():
    args = _parse_arguments()
    handle = _open(args)

    if interactive:
        print('.. Press ^C/^D to exit, or type hex bytes to write to the device.')

        import readline
        if args.history is None:
            import os.path
            args.history = os.path.join(os.path.expanduser('~'), '.hidconsole-history')
        try:
            readline.read_history_file(args.history)
        except Exception:
            # file may not exist yet
            pass

    try:
        from threading import Thread
        t = Thread(target=_continuous_read, args=(handle, ))
        t.daemon = True
        t.start()

        if interactive:
            # move the cursor at the bottom of the screen
            sys.stdout.write('\033[300B')  # move cusor at most 300 lines down, don't scroll

        while t.is_alive():
            line = read_packet(prompt)
            line = line.strip().replace(' ', '')
            # print ("line", line)
            if not line:
                continue

            data = _validate_input(line, args.hidpp)
            if data is None:
                continue

            _print('<<', data)
            _hid.write(handle, data)
            # wait for some kind of reply
            if args.hidpp and not interactive:
                rlist, wlist, xlist = _select([handle], [], [], 1)
                if data[1:2] == b'\xFF':
                    # the receiver will reply very fast, in a few milliseconds
                    time.sleep(0.010)
                else:
                    # the devices might reply quite slow
                    time.sleep(0.700)
    except EOFError:
        if interactive:
            print('')
        else:
            time.sleep(1)

    finally:
        print('.. Closing handle %r' % handle)
        _hid.close(handle)
        if interactive:
            readline.write_history_file(args.history)


if __name__ == '__main__':
    main()



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from time import time as _timestamp

from logitech_receiver import base as _base
from logitech_receiver import hidpp10 as _hidpp10
from logitech_receiver import notifications as _notifications
from logitech_receiver import status as _status


def run(receivers, args, find_receiver, _ignore):
    assert receivers

    if args.receiver:
        receiver_name = args.receiver.lower()
        receiver = find_receiver(receiver_name)
        if not receiver:
            raise Exception("no receiver found matching '%s'" % receiver_name)
    else:
        receiver = receivers[0]

    assert receiver
    receiver.status = _status.ReceiverStatus(receiver, lambda *args, **kwargs: None)

    # check if it's necessary to set the notification flags
    old_notification_flags = _hidpp10.get_notification_flags(receiver) or 0
    if not (old_notification_flags & _hidpp10.NOTIFICATION_FLAG.wireless):
        _hidpp10.set_notification_flags(receiver, old_notification_flags | _hidpp10.NOTIFICATION_FLAG.wireless)

    # get all current devices
    known_devices = [dev.number for dev in receiver]

    class _HandleWithNotificationHook(int):
        def notifications_hook(self, n):
            assert n
            if n.devnumber == 0xFF:
                _notifications.process(receiver, n)
            elif n.sub_id == 0x41:  # allow for other protocols! (was and n.address == 0x04)
                if n.devnumber not in known_devices:
                    receiver.status.new_device = receiver[n.devnumber]
                elif receiver.re_pairs:
                    del receiver[n.devnumber]  # get rid of information on device re-paired away
                    receiver.status.new_device = receiver[n.devnumber]

    timeout = 20  # seconds
    receiver.handle = _HandleWithNotificationHook(receiver.handle)

    receiver.set_lock(False, timeout=timeout)
    print('Pairing: turn your new device on (timing out in', timeout, 'seconds).')

    # the lock-open notification may come slightly later, wait for it a bit
    pairing_start = _timestamp()
    patience = 5  # seconds

    while receiver.status.lock_open or _timestamp() - pairing_start < patience:
        n = _base.read(receiver.handle)
        if n:
            n = _base.make_notification(*n)
            if n:
                receiver.handle.notifications_hook(n)

    if not (old_notification_flags & _hidpp10.NOTIFICATION_FLAG.wireless):
        # only clear the flags if they weren't set before, otherwise a
        # concurrently running Solaar app might stop working properly
        _hidpp10.set_notification_flags(receiver, old_notification_flags)

    if receiver.status.new_device:
        dev = receiver.status.new_device
        print('Paired device %d: %s (%s) [%s:%s]' % (dev.number, dev.name, dev.codename, dev.wpid, dev.serial))
    else:
        error = receiver.status.get(_status.KEYS.ERROR)
        if error:
            raise Exception('pairing failed: %s' % error)
        else:
            print('Paired a device')  # this is better than an error



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Translation support for the Logitech receivers library

from __future__ import absolute_import, division, print_function, unicode_literals

import gettext as _gettext

try:
    unicode  # noqa: F821
    _ = lambda x: _gettext.gettext(x).decode('UTF-8')
    ngettext = lambda *x: _gettext.ngettext(*x).decode('UTF-8')
except Exception:
    _ = _gettext.gettext
    ngettext = _gettext.ngettext

# A few common strings, not always accessible as such in the code.

_DUMMY = (
    # approximative battery levels
    _('empty'),
    _('critical'),
    _('low'),
    _('good'),
    _('full'),

    # battery charging statuses
    _('discharging'),
    _('recharging'),
    _('almost full'),
    _('charged'),
    _('slow recharge'),
    _('invalid battery'),
    _('thermal error'),

    # pairing errors
    _('device timeout'),
    _('device not supported'),
    _('too many devices'),
    _('sequence timeout'),

    # firmware kinds
    _('Firmware'),
    _('Bootloader'),
    _('Hardware'),
    _('Other'),
)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logging import DEBUG as _DEBUG
from logging import getLogger
from time import time as _timestamp

from . import hidpp10 as _hidpp10
from . import hidpp20 as _hidpp20
from .common import NamedInt as _NamedInt
from .common import NamedInts as _NamedInts
from .i18n import _, ngettext

_log = getLogger(__name__)
del getLogger

_R = _hidpp10.REGISTERS

#
#
#

ALERT = _NamedInts(NONE=0x00, NOTIFICATION=0x01, SHOW_WINDOW=0x02, ATTENTION=0x04, ALL=0xFF)

KEYS = _NamedInts(
    BATTERY_LEVEL=1,
    BATTERY_CHARGING=2,
    BATTERY_STATUS=3,
    LIGHT_LEVEL=4,
    LINK_ENCRYPTED=5,
    NOTIFICATION_FLAGS=6,
    ERROR=7,
    BATTERY_NEXT_LEVEL=8,
    BATTERY_VOLTAGE=9,
)

# If the battery charge is under this percentage, trigger an attention event
# (blink systray icon/notification/whatever).
_BATTERY_ATTENTION_LEVEL = 5

# If no updates have been receiver from the device for a while, ping the device
# and update it status accordingly.
# _STATUS_TIMEOUT = 5 * 60  # seconds
_LONG_SLEEP = 15 * 60  # seconds

#
#
#


def attach_to(device, changed_callback):
    assert device
    assert changed_callback

    if not hasattr(device, 'status') or device.status is None:
        if device.kind is None:
            device.status = ReceiverStatus(device, changed_callback)
        else:
            device.status = DeviceStatus(device, changed_callback)


#
#
#


class ReceiverStatus(dict):
    """The 'runtime' status of a receiver, mostly about the pairing process --
    is the pairing lock open or closed, any pairing errors, etc.
    """
    def __init__(self, receiver, changed_callback):
        assert receiver
        self._receiver = receiver

        assert changed_callback
        self._changed_callback = changed_callback

        # self.updated = 0

        self.lock_open = False
        self.new_device = None

        self[KEYS.ERROR] = None

    def __str__(self):
        count = len(self._receiver)
        return (
            _('No paired devices.')
            if count == 0 else ngettext('%(count)s paired device.', '%(count)s paired devices.', count) % {
                'count': count
            }
        )

    __unicode__ = __str__

    def changed(self, alert=ALERT.NOTIFICATION, reason=None):
        # self.updated = _timestamp()
        self._changed_callback(self._receiver, alert=alert, reason=reason)

    # def poll(self, timestamp):
    #     r = self._receiver
    #     assert r
    #
    #     if _log.isEnabledFor(_DEBUG):
    #         _log.debug("polling status of %s", r)
    #
    #     # make sure to read some stuff that may be read later by the UI
    #     r.serial, r.firmware, None
    #
    #     # get an update of the notification flags
    #     # self[KEYS.NOTIFICATION_FLAGS] = _hidpp10.get_notification_flags(r)


#
#
#


class DeviceStatus(dict):
    """Holds the 'runtime' status of a peripheral -- things like
    active/inactive, battery charge, lux, etc. It updates them mostly by
    processing incoming notification events from the device itself.
    """
    def __init__(self, device, changed_callback):
        assert device
        self._device = device

        assert changed_callback
        self._changed_callback = changed_callback

        # is the device active?
        self._active = None

        # timestamp of when this status object was last updated
        self.updated = 0

    def to_string(self):
        def _items():
            comma = False

            battery_level = self.get(KEYS.BATTERY_LEVEL)
            if battery_level is not None:
                if isinstance(battery_level, _NamedInt):
                    yield _('Battery: %(level)s') % {'level': _(str(battery_level))}
                else:
                    yield _('Battery: %(percent)d%%') % {'percent': battery_level}

                battery_status = self.get(KEYS.BATTERY_STATUS)
                if battery_status is not None:
                    yield ' (%s)' % _(str(battery_status))

                comma = True

            light_level = self.get(KEYS.LIGHT_LEVEL)
            if light_level is not None:
                if comma:
                    yield ', '
                yield _('Lighting: %(level)s lux') % {'level': light_level}

        return ''.join(i for i in _items())

    def __repr__(self):
        return '{' + ', '.join('\'%s\': %r' % (k, v) for k, v in self.items()) + '}'

    def __bool__(self):
        return bool(self._active)

    __nonzero__ = __bool__

    def set_battery_info(self, level, status, nextLevel=None, voltage=None, timestamp=None):
        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: battery %s, %s', self._device, level, status)

        if level is None:
            # Some notifications may come with no battery level info, just
            # charging state info, so do our best to infer a level (even if it is just the last level)
            # It is not always possible to do this well
            if status == _hidpp20.BATTERY_STATUS.full:
                level = _hidpp10.BATTERY_APPOX.full
            elif status in (_hidpp20.BATTERY_STATUS.almost_full, _hidpp20.BATTERY_STATUS.recharging):
                level = _hidpp10.BATTERY_APPOX.good
            elif status == _hidpp20.BATTERY_STATUS.slow_recharge:
                level = _hidpp10.BATTERY_APPOX.low
            else:
                level = self.get(KEYS.BATTERY_LEVEL)
        else:
            assert isinstance(level, int)

        # TODO: this is also executed when pressing Fn+F7 on K800.
        old_level, self[KEYS.BATTERY_LEVEL] = self.get(KEYS.BATTERY_LEVEL), level
        old_status, self[KEYS.BATTERY_STATUS] = self.get(KEYS.BATTERY_STATUS), status
        self[KEYS.BATTERY_NEXT_LEVEL] = nextLevel
        if voltage is not None:
            self[KEYS.BATTERY_VOLTAGE] = voltage

        charging = status in (
            _hidpp20.BATTERY_STATUS.recharging, _hidpp20.BATTERY_STATUS.almost_full, _hidpp20.BATTERY_STATUS.full,
            _hidpp20.BATTERY_STATUS.slow_recharge
        )
        old_charging, self[KEYS.BATTERY_CHARGING] = self.get(KEYS.BATTERY_CHARGING), charging

        changed = old_level != level or old_status != status or old_charging != charging
        alert, reason = ALERT.NONE, None

        if _hidpp20.BATTERY_OK(status) and (level is None or level > _BATTERY_ATTENTION_LEVEL):
            self[KEYS.ERROR] = None
        else:
            _log.warn('%s: battery %d%%, ALERT %s', self._device, level, status)
            if self.get(KEYS.ERROR) != status:
                self[KEYS.ERROR] = status
                # only show the notification once
                alert = ALERT.NOTIFICATION | ALERT.ATTENTION
            if isinstance(level, _NamedInt):
                reason = _('Battery: %(level)s (%(status)s)') % {'level': _(level), 'status': _(status)}
            else:
                reason = _('Battery: %(percent)d%% (%(status)s)') % {'percent': level, 'status': status.name}

        if changed or reason:
            # update the leds on the device, if any
            _hidpp10.set_3leds(self._device, level, charging=charging, warning=bool(alert))
            self.changed(active=True, alert=alert, reason=reason, timestamp=timestamp)

    # Retrieve and regularize battery status
    def read_battery(self, timestamp=None):
        if self._active:
            d = self._device
            assert d

            if d.protocol < 2.0:
                battery = _hidpp10.get_battery(d)
                self.set_battery_keys(battery)
                return

            battery = _hidpp20.get_battery(d)
            if battery is None:
                v = _hidpp20.get_voltage(d)
                if v is not None:
                    level, status, voltage, _ignore, _ignore = v
                    self.set_battery_keys((level, status, None), voltage)
                    return

            # Really unnecessary, if the device has SOLAR_DASHBOARD it should be
            # broadcasting it's battery status anyway, it will just take a little while.
            # However, when the device has just been detected, it will not show
            # any battery status for a while (broadcasts happen every 90 seconds).
            if battery is None and d.features and _hidpp20.FEATURE.SOLAR_DASHBOARD in d.features:
                d.feature_request(_hidpp20.FEATURE.SOLAR_DASHBOARD, 0x00, 1, 1)
                return
            self.set_battery_keys(battery)

    def set_battery_keys(self, battery, voltage=None):
        if battery is not None:
            level, status, nextLevel = battery
            self.set_battery_info(level, status, nextLevel, voltage)
        elif KEYS.BATTERY_STATUS in self:
            self[KEYS.BATTERY_STATUS] = None
            self[KEYS.BATTERY_CHARGING] = None
            self.changed()

    def changed(self, active=None, alert=ALERT.NONE, reason=None, timestamp=None):
        assert self._changed_callback
        d = self._device
        # assert d  # may be invalid when processing the 'unpaired' notification
        timestamp = timestamp or _timestamp()

        if active is not None:
            d.online = active
            was_active, self._active = self._active, active
            if active:
                if not was_active:
                    # Make sure to set notification flags on the device, they
                    # get cleared when the device is turned off (but not when the device
                    # goes idle, and we can't tell the difference right now).
                    if d.protocol < 2.0:
                        self[KEYS.NOTIFICATION_FLAGS] = d.enable_notifications()

                    # If we've been inactive for a long time, forget anything
                    # about the battery.
                    if self.updated > 0 and timestamp - self.updated > _LONG_SLEEP:
                        self[KEYS.BATTERY_LEVEL] = None
                        self[KEYS.BATTERY_STATUS] = None
                        self[KEYS.BATTERY_CHARGING] = None

                    # Devices lose configuration when they are turned off,
                    # make sure they're up-to-date.
                    if _log.isEnabledFor(_DEBUG):
                        _log.debug('%s pushing device settings %s', d, d.settings)
                    for s in d.settings:
                        s.apply()

                    if self.get(KEYS.BATTERY_LEVEL) is None:
                        self.read_battery(timestamp)
            else:
                if was_active:
                    battery = self.get(KEYS.BATTERY_LEVEL)
                    self.clear()
                    # If we had a known battery level before, assume it's not going
                    # to change much while the device is offline.
                    if battery is not None:
                        self[KEYS.BATTERY_LEVEL] = battery

        if self.updated == 0 and active is True:
            # if the device is active on the very first status notification,
            # (meaning just when the program started or a new receiver was just
            # detected), pop-up a notification about it
            alert |= ALERT.NOTIFICATION
        self.updated = timestamp

        # if _log.isEnabledFor(_DEBUG):
        #     _log.debug("device %d changed: active=%s %s", d.number, self._active, dict(self))
        self._changed_callback(d, alert, reason)

    # def poll(self, timestamp):
    #     d = self._device
    #     if not d:
    #         _log.error("polling status of invalid device")
    #         return
    #
    #     if self._active:
    #         if _log.isEnabledFor(_DEBUG):
    #             _log.debug("polling status of %s", d)
    #
    #         # read these from the device, the UI may need them later
    #         d.protocol, d.serial, d.firmware, d.kind, d.name, d.settings, None
    #
    #         # make sure we know all the features of the device
    #         # if d.features:
    #         #     d.features[:]
    #
    #         # devices may go out-of-range while still active, or the computer
    #         # may go to sleep and wake up without the devices available
    #         if timestamp - self.updated > _STATUS_TIMEOUT:
    #             if d.ping():
    #                 timestamp = self.updated = _timestamp()
    #             else:
    #                 self.changed(active=False, reason='out of range')
    #
    #         # if still active, make sure we know the battery level
    #         if KEYS.BATTERY_LEVEL not in self:
    #             self.read_battery(timestamp)
    #
    #     elif timestamp - self.updated > _STATUS_TIMEOUT:
    #         if d.ping():
    #             self.changed(active=True)
    #         else:
    #             self.updated = _timestamp()



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2020
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logitech_receiver import hidpp10 as _hidpp10
from logitech_receiver.common import strhex as _strhex
from solaar.cli.show import _print_receiver

_R = _hidpp10.REGISTERS


def run(receivers, args, find_receiver, _ignore):
    assert receivers

    if args.receiver:
        receiver_name = args.receiver.lower()
        receiver = find_receiver(receiver_name)
        if not receiver:
            raise Exception("no receiver found matching '%s'" % receiver_name)
    else:
        receiver = receivers[0]

    assert receiver

    _print_receiver(receiver)

    print('  Register Dump')
    register = receiver.read_register(_R.notifications)
    print('    Notification Register %#04x: %s' % (_R.notifications % 0x100, '0x' + _strhex(register) if register else 'None'))
    register = receiver.read_register(_R.receiver_connection)
    print(
        '    Connection State      %#04x: %s' %
        (_R.receiver_connection % 0x100, '0x' + _strhex(register) if register else 'None')
    )
    register = receiver.read_register(_R.devices_activity)
    print(
        '    Device Activity       %#04x: %s' %
        (_R.devices_activity % 0x100, '0x' + _strhex(register) if register else 'None')
    )

    for device in range(0, 6):
        for sub_reg in [0x0, 0x10, 0x20, 0x30]:
            register = receiver.read_register(_R.receiver_info, sub_reg + device)
            print(
                '    Pairing Register %#04x %#04x: %s' %
                (_R.receiver_info % 0x100, sub_reg + device, '0x' + _strhex(register) if register else 'None')
            )
        register = receiver.read_register(_R.receiver_info, 0x40 + device)
        print(
            '    Pairing Name     %#04x %#02x: %s' %
            (_R.receiver_info % 0x100, 0x40 + device, register[2:2 + ord(register[1:2])] if register else 'None')
        )

    for sub_reg in range(0, 5):
        register = receiver.read_register(_R.firmware, sub_reg)
        print(
            '    Firmware         %#04x %#04x: %s' %
            (_R.firmware % 0x100, sub_reg, '0x' + _strhex(register) if register else 'None')
        )



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logitech_receiver import hidpp10 as _hidpp10
from logitech_receiver import hidpp20 as _hidpp20
from logitech_receiver import settings_templates as _settings_templates
from logitech_receiver import special_keys as _special_keys
from logitech_receiver.common import NamedInt as _NamedInt


def _print_receiver(receiver):
    paired_count = receiver.count()

    print(receiver.name)
    print('  Device path  :', receiver.path)
    print('  USB id       : 046d:%s' % receiver.product_id)
    print('  Serial       :', receiver.serial)
    if receiver.firmware:
        for f in receiver.firmware:
            print('    %-11s: %s' % (f.kind, f.version))

    print('  Has', paired_count, 'paired device(s) out of a maximum of %d.' % receiver.max_devices)
    if receiver.remaining_pairings() and receiver.remaining_pairings() >= 0:
        print('  Has %d successful pairing(s) remaining.' % receiver.remaining_pairings())

    notification_flags = _hidpp10.get_notification_flags(receiver)
    if notification_flags is not None:
        if notification_flags:
            notification_names = _hidpp10.NOTIFICATION_FLAG.flag_names(notification_flags)
            print('  Notifications: %s (0x%06X)' % (', '.join(notification_names), notification_flags))
        else:
            print('  Notifications: (none)')

    activity = receiver.read_register(_hidpp10.REGISTERS.devices_activity)
    if activity:
        activity = [(d, ord(activity[d - 1:d])) for d in range(1, receiver.max_devices)]
        activity_text = ', '.join(('%d=%d' % (d, a)) for d, a in activity if a > 0)
        print('  Device activity counters:', activity_text or '(empty)')


def _battery_text(level):
    if level is None:
        return 'N/A'
    elif isinstance(level, _NamedInt):
        return str(level)
    else:
        return '%d%%' % level


def _print_device(dev):
    assert dev
    # check if the device is online
    dev.ping()

    print('  %d: %s' % (dev.number, dev.name))
    print('     Codename     :', dev.codename)
    print('     Kind         :', dev.kind)
    print('     Wireless PID :', dev.wpid)
    if dev.protocol:
        print('     Protocol     : HID++ %1.1f' % dev.protocol)
    else:
        print('     Protocol     : unknown (device is offline)')
    if dev.polling_rate:
        print('     Polling rate :', dev.polling_rate, 'ms (%dHz)' % (1000 // dev.polling_rate))
    print('     Serial number:', dev.serial)
    if dev.firmware:
        for fw in dev.firmware:
            print('       %11s:' % fw.kind, (fw.name + ' ' + fw.version).strip())

    if dev.power_switch_location:
        print('     The power switch is located on the %s.' % dev.power_switch_location)

    if dev.online:
        notification_flags = _hidpp10.get_notification_flags(dev)
        if notification_flags is not None:
            if notification_flags:
                notification_names = _hidpp10.NOTIFICATION_FLAG.flag_names(notification_flags)
                print('     Notifications: %s (0x%06X).' % (', '.join(notification_names), notification_flags))
            else:
                print('     Notifications: (none).')
        device_features = _hidpp10.get_device_features(dev)
        if device_features is not None:
            if device_features:
                device_features_names = _hidpp10.DEVICE_FEATURES.flag_names(device_features)
                print('     Features: %s (0x%06X)' % (', '.join(device_features_names), device_features))
            else:
                print('     Features: (none)')

    if dev.online and dev.features:
        print('     Supports %d HID++ 2.0 features:' % len(dev.features))
        dev.persister = None  # Give the device a fake persister
        dev_settings = []
        _settings_templates.check_feature_settings(dev, dev_settings)
        for index, feature in enumerate(dev.features):
            feature = dev.features[index]
            flags = dev.request(0x0000, feature.bytes(2))
            flags = 0 if flags is None else ord(flags[1:2])
            flags = _hidpp20.FEATURE_FLAG.flag_names(flags)
            print('        %2d: %-22s {%04X}   %s' % (index, feature, feature, ', '.join(flags)))
            if feature == _hidpp20.FEATURE.HIRES_WHEEL:
                wheel = _hidpp20.get_hires_wheel(dev)
                if wheel:
                    multi, has_invert, has_switch, inv, res, target, ratchet = wheel
                    print('            Multiplier: %s' % multi)
                    if has_invert:
                        print('            Has invert')
                        if inv:
                            print('              Inverse wheel motion')
                        else:
                            print('              Normal wheel motion')
                    if has_switch:
                        print('            Has ratchet switch')
                        if ratchet:
                            print('              Normal wheel mode')
                        else:
                            print('              Free wheel mode')
                    if res:
                        print('            High resolution mode')
                    else:
                        print('            Low resolution mode')
                    if target:
                        print('            HID++ notification')
                    else:
                        print('            HID notification')
            elif feature == _hidpp20.FEATURE.MOUSE_POINTER:
                mouse_pointer = _hidpp20.get_mouse_pointer_info(dev)
                if mouse_pointer:
                    print('            DPI: %s' % mouse_pointer['dpi'])
                    print('            Acceleration: %s' % mouse_pointer['acceleration'])
                    if mouse_pointer['suggest_os_ballistics']:
                        print('            Use OS ballistics')
                    else:
                        print('            Override OS ballistics')
                    if mouse_pointer['suggest_vertical_orientation']:
                        print('            Provide vertical tuning, trackball')
                    else:
                        print('            No vertical tuning, standard mice')
            if feature == _hidpp20.FEATURE.VERTICAL_SCROLLING:
                vertical_scrolling_info = _hidpp20.get_vertical_scrolling_info(dev)
                if vertical_scrolling_info:
                    print('            Roller type: %s' % vertical_scrolling_info['roller'])
                    print('            Ratchet per turn: %s' % vertical_scrolling_info['ratchet'])
                    print('            Scroll lines: %s' % vertical_scrolling_info['lines'])
            elif feature == _hidpp20.FEATURE.HI_RES_SCROLLING:
                scrolling_mode, scrolling_resolution = _hidpp20.get_hi_res_scrolling_info(dev)
                if scrolling_mode:
                    print('            Hi-res scrolling enabled')
                else:
                    print('            Hi-res scrolling disabled')
                if scrolling_resolution:
                    print('            Hi-res scrolling multiplier: %s' % scrolling_resolution)
            elif feature == _hidpp20.FEATURE.POINTER_SPEED:
                pointer_speed = _hidpp20.get_pointer_speed_info(dev)
                if pointer_speed:
                    print('            Pointer Speed: %s' % pointer_speed)
            elif feature == _hidpp20.FEATURE.LOWRES_WHEEL:
                wheel_status = _hidpp20.get_lowres_wheel_status(dev)
                if wheel_status:
                    print('            Wheel Reports: %s' % wheel_status)
            elif feature == _hidpp20.FEATURE.NEW_FN_INVERSION:
                inverted, default_inverted = _hidpp20.get_new_fn_inversion(dev)
                print('            Fn-swap:', 'enabled' if inverted else 'disabled')
                print('            Fn-swap default:', 'enabled' if default_inverted else 'disabled')
            elif feature == _hidpp20.FEATURE.HOSTS_INFO:
                host_names = _hidpp20.get_host_names(dev)
                for host, (paired, name) in host_names.items():
                    print('            Host %s (%s): %s' % (host, 'paired' if paired else 'unpaired', name))
            for setting in dev_settings:
                if setting.feature == feature:
                    v = setting.read(False)
                    print('            %s: %s' % (setting.label, v))

    if dev.online and dev.keys:
        print('     Has %d reprogrammable keys:' % len(dev.keys))
        for k in dev.keys:
            flags = _special_keys.KEY_FLAG.flag_names(k.flags)
            # TODO: add here additional variants for other REPROG_CONTROLS
            if dev.keys.keyversion == 1:
                print('        %2d: %-26s => %-27s   %s' % (k.index, k.key, k.task, ', '.join(flags)))
            if dev.keys.keyversion == 4:
                print('        %2d: %-26s, default: %-27s => %-26s' % (k.index, k.key, k.task, k.remapped))
                print('             %s, pos:%d, group:%1d, gmask:%d' % (', '.join(flags), k.pos, k.group, k.group_mask))
    if dev.online:
        battery = _hidpp20.get_battery(dev)
        if battery is None:
            battery = _hidpp10.get_battery(dev)
        if battery is not None:
            level, status, nextLevel = battery
            text = _battery_text(level)
            nextText = '' if nextLevel is None else ', next level ' + _battery_text(nextLevel)
            print('     Battery: %s, %s%s.' % (text, status, nextText))
        else:
            battery_voltage = _hidpp20.get_voltage(dev)
            if battery_voltage:
                (level, status, voltage, charge_sts, charge_type) = battery_voltage
                print('     Battery: %smV, %s, %s.' % (voltage, status, level))
            else:
                print('     Battery status unavailable.')
    else:
        print('     Battery: unknown (device is offline).')


def run(receivers, args, find_receiver, find_device):
    assert receivers
    assert args.device

    device_name = args.device.lower()

    if device_name == 'all':
        for r in receivers:
            _print_receiver(r)
            count = r.count()
            if count:
                for dev in r:
                    print('')
                    _print_device(dev)
                    count -= 1
                    if not count:
                        break
            print('')
        return

    dev = find_receiver(receivers, device_name)
    if dev:
        _print_receiver(dev)
        return

    dev = find_device(receivers, device_name)
    assert dev
    _print_device(dev)



#!/usr/bin/env python3
# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

import importlib
import os.path

from logging import INFO as _INFO
from logging import getLogger

import solaar.cli as _cli
import solaar.i18n as _i18n

from solaar import NAME, __version__

_log = getLogger(__name__)
del getLogger

#
#
#


def _require(module, os_package, gi=None, gi_package=None, gi_version=None):
    try:
        if gi is not None:
            gi.require_version(gi_package, gi_version)
        return importlib.import_module(module)
    except (ImportError, ValueError):
        import sys
        sys.exit('%s: missing required system package %s' % (NAME, os_package))


prefer_symbolic_battery_icons = False


def _parse_arguments():
    import argparse
    arg_parser = argparse.ArgumentParser(prog=NAME.lower())
    arg_parser.add_argument(
        '-d',
        '--debug',
        action='count',
        default=0,
        help='print logging messages, for debugging purposes (may be repeated for extra verbosity)'
    )
    arg_parser.add_argument(
        '-D',
        '--hidraw',
        action='store',
        dest='hidraw_path',
        metavar='PATH',
        help='unifying receiver to use; the first detected receiver if unspecified. Example: /dev/hidraw2'
    )
    arg_parser.add_argument('--restart-on-wake-up', action='store_true', help='restart Solaar on sleep wake-up (experimental)')
    arg_parser.add_argument(
        '-w', '--window', choices=('show', 'hide', 'only'), help='start with window showing / hidden / only (no tray icon)'
    )
    arg_parser.add_argument('-b', '--battery-icons', choices=('regular', 'symbolic'), help='prefer regular / symbolic icons')
    arg_parser.add_argument('-V', '--version', action='version', version='%(prog)s ' + __version__)
    arg_parser.add_argument('--help-actions', action='store_true', help='print help for the optional actions')
    arg_parser.add_argument('action', nargs=argparse.REMAINDER, choices=_cli.actions, help='optional actions to perform')

    args = arg_parser.parse_args()

    if args.help_actions:
        _cli.print_help()
        return

    if args.window is None:
        args.window = 'show'  # default behaviour is to show main window

    global prefer_symbolic_battery_icons
    prefer_symbolic_battery_icons = True if args.battery_icons == 'symbolic' else False

    import logging
    if args.debug > 0:
        log_level = logging.WARNING - 10 * args.debug
        log_format = '%(asctime)s,%(msecs)03d %(levelname)8s [%(threadName)s] %(name)s: %(message)s'
        logging.basicConfig(level=max(log_level, logging.DEBUG), format=log_format, datefmt='%H:%M:%S')
    else:
        logging.root.addHandler(logging.NullHandler())
        logging.root.setLevel(logging.ERROR)

    if not args.action:
        if logging.root.isEnabledFor(logging.INFO):
            logging.info('language %s (%s), translations path %s', _i18n.language, _i18n.encoding, _i18n.path)

    return args


# On first SIGINT, dump threads to stderr; on second, exit
def _handlesigint(signal, stack):
    import signal
    import sys
    import faulthandler
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    if _log.isEnabledFor(_INFO):
        faulthandler.dump_traceback()

    sys.exit('%s: exit due to keyboard interrupt' % (NAME.lower()))


def main():
    _require('pyudev', 'python3-pyudev')

    # handle ^C in console
    import signal
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGINT, _handlesigint)

    args = _parse_arguments()
    if not args:
        return
    if args.action:
        # if any argument, run comandline and exit
        return _cli.run(args.action, args.hidraw_path)

    gi = _require('gi', 'python3-gi or python3-gobject')
    _require('gi.repository.Gtk', 'gir1.2-gtk-3.0', gi, 'Gtk', '3.0')

    udev_file = '42-logitech-unify-permissions.rules'
    if not os.path.isfile('/etc/udev/rules.d/' + udev_file) and not os.path.isfile('/lib/udev/rules.d/' + udev_file):
        print(_i18n._('Solaar depends on a udev file that is not present'))
        print(
            _i18n._(
                'For more information see the Solaar installation directions\n'
                'at https://pwr-solaar.github.io/Solaar/installation'
            )
        )

    try:
        import solaar.ui as ui
        import solaar.listener as listener
        listener.setup_scanner(ui.status_changed, ui.error_dialog)

        import solaar.upower as _upower
        if args.restart_on_wake_up:
            _upower.watch(listener.start_all, listener.stop_all)
        else:
            _upower.watch(lambda: listener.ping_all(True))

        # main UI event loop
        ui.run_loop(listener.start_all, listener.stop_all, args.window != 'only', args.window != 'hide')
    except Exception:
        import sys
        from traceback import format_exc
        sys.exit('%s: error: %s' % (NAME.lower(), format_exc()))


if __name__ == '__main__':
    main()



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logitech_receiver import settings as _settings
from logitech_receiver import settings_templates as _settings_templates
from solaar import configuration as _configuration


def _print_setting(s, verbose=True):
    print('#', s.label)
    if verbose:
        if s.description:
            print('#', s.description.replace('\n', ' '))
        if s.kind == _settings.KIND.toggle:
            print('#   possible values: on/true/t/yes/y/1 or off/false/f/no/n/0')
        elif s.kind == _settings.KIND.choice:
            print(
                '#   possible values: one of [', ', '.join(str(v) for v in s.choices),
                '], or higher/lower/highest/max/lowest/min'
            )
        else:
            # wtf?
            pass
    value = s.read(cached=False)
    if value is None:
        print(s.name, '= ? (failed to read from device)')
    else:
        print(s.name, '= %r' % value)


def run(receivers, args, find_receiver, find_device):
    assert receivers
    assert args.device

    device_name = args.device.lower()
    dev = find_device(receivers, device_name)

    if not dev.ping():
        raise Exception('%s is offline' % dev.name)

    if not args.setting:  # print all settings, so first set them all up
        if not dev.settings:
            raise Exception('no settings for %s' % dev.name)
        _configuration.attach_to(dev)
        for s in dev.settings:
            s.apply()
        print(dev.name, '(%s) [%s:%s]' % (dev.codename, dev.wpid, dev.serial))
        for s in dev.settings:
            print('')
            _print_setting(s)
        return

    setting_name = args.setting.lower()
    setting = _settings_templates.check_feature_setting(dev, setting_name)
    if setting is None:
        raise Exception("no setting '%s' for %s" % (args.setting, dev.name))
    _configuration.attach_to(dev)

    if args.value is None:
        setting.apply()
        _print_setting(setting)
        return

    if setting.kind == _settings.KIND.toggle:
        value = args.value
        try:
            value = bool(int(value))
        except Exception:
            if value.lower() in ('true', 'yes', 'on', 't', 'y'):
                value = True
            elif value.lower() in ('false', 'no', 'off', 'f', 'n'):
                value = False
            else:
                raise Exception("%s: don't know how to interpret '%s' as boolean" % (setting.name, value))
        print('Setting %s to %s' % (setting.name, value))

    elif setting.kind == _settings.KIND.range:
        try:
            value = int(args.value)
        except ValueError:
            raise Exception("%s: can't interpret '%s' as integer" % (setting.name, args.value))
        min, max = setting.range
        if value < min or value > max:
            raise Exception("%s: value '%s' out of bounds" % (setting.name, args.value))
        print('Setting %s to %s' % (setting.name, value))

    elif setting.kind == _settings.KIND.choice:
        value = args.value
        lvalue = value.lower()
        try:
            ivalue = int(value)
        except ValueError:
            ivalue = None
        if value in setting.choices:
            value = setting.choices[value]
        elif ivalue is not None and ivalue >= 0 and ivalue < len(setting.choices):
            value = setting.choices[ivalue]
        elif lvalue in ('higher', 'lower'):
            old_value = setting.read()
            if old_value is None:
                raise Exception("%s: could not read current value'" % setting.name)
            if lvalue == 'lower':
                lower_values = setting.choices[:old_value]
                value = lower_values[-1] if lower_values else setting.choices[:][0]
            elif lvalue == 'higher':
                higher_values = setting.choices[old_value + 1:]
                value = higher_values[0] if higher_values else setting.choices[:][-1]
        elif lvalue in ('highest', 'max', 'first'):
            value = setting.choices[:][-1]
        elif lvalue in ('lowest', 'min', 'last'):
            value = setting.choices[:][0]
        else:
            raise Exception('%s: possible values are [%s]' % (setting.name, ', '.join(str(v) for v in setting.choices)))
        print('Setting %s to %r' % (setting.name, value))

    else:
        raise Exception('NotImplemented')

    result = setting.write(value)
    if result is None:
        raise Exception("%s: failed to set value '%s' [%r]" % (setting.name, str(value), value))



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os as _os
import os.path as _path

from json import dump as _json_save
from json import load as _json_load
from logging import DEBUG as _DEBUG
from logging import INFO as _INFO
from logging import getLogger

from solaar import __version__

_log = getLogger(__name__)
del getLogger

_XDG_CONFIG_HOME = _os.environ.get('XDG_CONFIG_HOME') or _path.expanduser(_path.join('~', '.config'))
_file_path = _path.join(_XDG_CONFIG_HOME, 'solaar', 'config.json')

_KEY_VERSION = '_version'
_KEY_NAME = '_name'
_configuration = {}


def _load():
    if _path.isfile(_file_path):
        loaded_configuration = {}
        try:
            with open(_file_path, 'r') as config_file:
                loaded_configuration = _json_load(config_file)
        except Exception:
            _log.error('failed to load from %s', _file_path)

        # loaded_configuration.update(_configuration)
        _configuration.clear()
        _configuration.update(loaded_configuration)

    if _log.isEnabledFor(_DEBUG):
        _log.debug('load => %s', _configuration)

    _cleanup(_configuration)
    _configuration[_KEY_VERSION] = __version__
    return _configuration


def save():
    # don't save if the configuration hasn't been loaded
    if _KEY_VERSION not in _configuration:
        return

    dirname = _os.path.dirname(_file_path)
    if not _path.isdir(dirname):
        try:
            _os.makedirs(dirname)
        except Exception:
            _log.error('failed to create %s', dirname)
            return False

    _cleanup(_configuration)

    try:
        with open(_file_path, 'w') as config_file:
            _json_save(_configuration, config_file, skipkeys=True, indent=2, sort_keys=True)

        if _log.isEnabledFor(_INFO):
            _log.info('saved %s to %s', _configuration, _file_path)
        return True
    except Exception:
        _log.error('failed to save to %s', _file_path)


def _cleanup(d):
    # remove None values from the dict
    for key in list(d.keys()):
        value = d.get(key)
        if value is None:
            del d[key]
        elif isinstance(value, dict):
            _cleanup(value)


def _device_key(device):
    return '%s:%s' % (device.wpid, device.serial)


class _DeviceEntry(dict):
    def __init__(self, *args, **kwargs):
        super(_DeviceEntry, self).__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        super(_DeviceEntry, self).__setitem__(key, value)
        save()


def _device_entry(device):
    if not _configuration:
        _load()

    device_key = _device_key(device)
    c = _configuration.get(device_key) or {}

    if not isinstance(c, _DeviceEntry):
        c[_KEY_NAME] = device.name
        c = _DeviceEntry(c)
        _configuration[device_key] = c

    return c


def attach_to(device):
    """Apply the last saved configuration to a device."""
    if not _configuration:
        _load()

    persister = _device_entry(device)
    device.persister = persister



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Handles incoming events from the receiver/devices, updating the related
# status object as appropriate.

from __future__ import absolute_import, division, print_function, unicode_literals

from logging import DEBUG as _DEBUG
from logging import INFO as _INFO
from logging import getLogger

from . import hidpp10 as _hidpp10
from . import hidpp20 as _hidpp20
from .base import DJ_NOTIFICATION_LENGTH as _DJ_NOTIFICATION_LENGTH
from .common import strhex as _strhex
from .common import unpack as _unpack
from .i18n import _
from .status import ALERT as _ALERT
from .status import KEYS as _K

_log = getLogger(__name__)
del getLogger

_R = _hidpp10.REGISTERS
_F = _hidpp20.FEATURE

#
#
#


def process(device, notification):
    assert device
    assert notification

    assert hasattr(device, 'status')
    status = device.status
    assert status is not None

    if device.kind is None:
        return _process_receiver_notification(device, status, notification)

    return _process_device_notification(device, status, notification)


#
#
#


def _process_receiver_notification(receiver, status, n):
    # supposedly only 0x4x notifications arrive for the receiver
    assert n.sub_id & 0x40 == 0x40

    # pairing lock notification
    if n.sub_id == 0x4A:
        status.lock_open = bool(n.address & 0x01)
        reason = (_('pairing lock is open') if status.lock_open else _('pairing lock is closed'))
        if _log.isEnabledFor(_INFO):
            _log.info('%s: %s', receiver, reason)

        status[_K.ERROR] = None
        if status.lock_open:
            status.new_device = None

        pair_error = ord(n.data[:1])
        if pair_error:
            status[_K.ERROR] = error_string = _hidpp10.PAIRING_ERRORS[pair_error]
            status.new_device = None
            _log.warn('pairing error %d: %s', pair_error, error_string)

        status.changed(reason=reason)
        return True

    _log.warn('%s: unhandled notification %s', receiver, n)


#
#
#


def _process_device_notification(device, status, n):
    # incoming packets with SubId >= 0x80 are supposedly replies from
    # HID++ 1.0 requests, should never get here
    assert n.sub_id & 0x80 == 0

    # 0x40 to 0x7F appear to be HID++ 1.0 or DJ notifications
    if n.sub_id >= 0x40:
        if len(n.data) == _DJ_NOTIFICATION_LENGTH:
            return _process_dj_notification(device, status, n)
        else:
            return _process_hidpp10_notification(device, status, n)

    # At this point, we need to know the device's protocol, otherwise it's
    # possible to not know how to handle it.
    assert device.protocol is not None

    # some custom battery events for HID++ 1.0 devices
    if device.protocol < 2.0:
        return _process_hidpp10_custom_notification(device, status, n)

    # assuming 0x00 to 0x3F are feature (HID++ 2.0) notifications
    assert device.features
    try:
        feature = device.features[n.sub_id]
    except IndexError:
        _log.warn('%s: notification from invalid feature index %02X: %s', device, n.sub_id, n)
        return False

    return _process_feature_notification(device, status, n, feature)


def _process_dj_notification(device, status, n):
    if _log.isEnabledFor(_DEBUG):
        _log.debug('%s (%s) DJ notification %s', device, device.protocol, n)

    if n.sub_id == 0x40:
        # do all DJ paired notifications also show up as HID++ 1.0 notifications?
        if _log.isEnabledFor(_INFO):
            _log.info('%s: ignoring DJ unpaired: %s', device, n)
        return True

    if n.sub_id == 0x41:
        # do all DJ paired notifications also show up as HID++ 1.0 notifications?
        if _log.isEnabledFor(_INFO):
            _log.info('%s: ignoring DJ paired: %s', device, n)
        return True

    if n.sub_id == 0x42:
        if _log.isEnabledFor(_INFO):
            _log.info('%s: ignoring DJ connection: %s', device, n)
        return True

    _log.warn('%s: unrecognized DJ %s', device, n)


def _process_hidpp10_custom_notification(device, status, n):
    if _log.isEnabledFor(_DEBUG):
        _log.debug('%s (%s) custom notification %s', device, device.protocol, n)

    if n.sub_id in (_R.battery_status, _R.battery_charge):
        # message layout: 10 ix <register> <xx> <yy> <zz> <00>
        assert n.data[-1:] == b'\x00'
        data = chr(n.address).encode() + n.data
        charge, status_text, next_charge = _hidpp10.parse_battery_status(n.sub_id, data)
        status.set_battery_info(charge, status_text, next_charge)
        return True

    if n.sub_id == _R.keyboard_illumination:
        # message layout: 10 ix 17("address")  <??> <?> <??> <light level 1=off..5=max>
        # TODO anything we can do with this?
        if _log.isEnabledFor(_INFO):
            _log.info('illumination event: %s', n)
        return True

    _log.warn('%s: unrecognized %s', device, n)


def _process_hidpp10_notification(device, status, n):
    # unpair notification
    if n.sub_id == 0x40:
        if n.address == 0x02:
            # device un-paired
            status.clear()
            device.wpid = None
            device.status = None
            if device.number in device.receiver:
                del device.receiver[device.number]
            status.changed(active=False, alert=_ALERT.ALL, reason=_('unpaired'))
        else:
            _log.warn('%s: disconnection with unknown type %02X: %s', device, n.address, n)
        return True

    # wireless link notification
    if n.sub_id == 0x41:
        protocol_name = (
            'Bluetooth' if n.address == 0x01 else '27 MHz' if n.address == 0x02 else
            'QUAD or eQUAD' if n.address == 0x03 else 'eQUAD step 4 DJ' if n.address == 0x04 else 'DFU Lite' if n.address ==
            0x05 else 'eQUAD step 4 Lite' if n.address == 0x06 else 'eQUAD step 4 Gaming' if n.address ==
            0x07 else 'eQUAD step 4 for gamepads' if n.address == 0x08 else 'eQUAD nano Lite' if n.address ==
            0x0A else 'Lightspeed 1' if n.address == 0x0C else 'Lightspeed 1_1' if n.address == 0x0D else None
        )
        if protocol_name:
            if _log.isEnabledFor(_DEBUG):
                wpid = _strhex(n.data[2:3] + n.data[1:2])
                # workaround for short EX100 wpids
                if protocol_name == '27 MHz':
                    wpid = _strhex(n.data[2:3]) + '00'
                assert wpid == device.wpid, '%s wpid mismatch, got %s' % (device, wpid)

            flags = ord(n.data[:1]) & 0xF0
            link_encrypted = bool(flags & 0x20)
            link_established = not (flags & 0x40)
            if _log.isEnabledFor(_DEBUG):
                sw_present = bool(flags & 0x10)
                has_payload = bool(flags & 0x80)
                _log.debug(
                    '%s: %s connection notification: software=%s, encrypted=%s, link=%s, payload=%s', device, protocol_name,
                    sw_present, link_encrypted, link_established, has_payload
                )
            status[_K.LINK_ENCRYPTED] = link_encrypted
            status.changed(active=link_established)
        else:
            _log.warn('%s: connection notification with unknown protocol %02X: %s', device.number, n.address, n)

        return True

    if n.sub_id == 0x49:
        # raw input event? just ignore it
        # if n.address == 0x01, no idea what it is, but they keep on coming
        # if n.address == 0x03, appears to be an actual input event,
        #     because they only come when input happents
        return True

    # power notification
    if n.sub_id == 0x4B:
        if n.address == 0x01:
            if _log.isEnabledFor(_DEBUG):
                _log.debug('%s: device powered on', device)
            reason = status.to_string() or _('powered on')
            status.changed(active=True, alert=_ALERT.NOTIFICATION, reason=reason)
        else:
            _log.warn('%s: unknown %s', device, n)
        return True

    _log.warn('%s: unrecognized %s', device, n)


def _process_feature_notification(device, status, n, feature):
    if feature == _F.BATTERY_STATUS:
        if n.address == 0x00:
            discharge_level = ord(n.data[:1])
            discharge_level = None if discharge_level == 0 else discharge_level
            discharge_next_level = ord(n.data[1:2])
            battery_status = ord(n.data[2:3])
            status.set_battery_info(discharge_level, _hidpp20.BATTERY_STATUS[battery_status], discharge_next_level)
        else:
            _log.warn('%s: unknown BATTERY %s', device, n)
        return True

    if feature == _F.BATTERY_VOLTAGE:
        if n.address == 0x00:
            level, status, voltage, _ignore, _ignore = _hidpp20.decipher_voltage(n.data)
            status.set_battery_info(level, status, None, voltage)
        else:
            _log.warn('%s: unknown VOLTAGE %s', device, n)
        return True

    # TODO: what are REPROG_CONTROLS_V{2,3}?
    if feature == _F.REPROG_CONTROLS:
        if n.address == 0x00:
            if _log.isEnabledFor(_INFO):
                _log.info('%s: reprogrammable key: %s', device, n)
        else:
            _log.warn('%s: unknown REPROGRAMMABLE KEYS %s', device, n)
        return True

    if feature == _F.WIRELESS_DEVICE_STATUS:
        if n.address == 0x00:
            if _log.isEnabledFor(_DEBUG):
                _log.debug('wireless status: %s', n)
            if n.data[0:3] == b'\x01\x01\x01':
                status.changed(active=True, alert=_ALERT.NOTIFICATION, reason='powered on')
            else:
                _log.warn('%s: unknown WIRELESS %s', device, n)
        else:
            _log.warn('%s: unknown WIRELESS %s', device, n)
        return True

    if feature == _F.SOLAR_DASHBOARD:
        if n.data[5:9] == b'GOOD':
            charge, lux, adc = _unpack('!BHH', n.data[:5])
            # guesstimate the battery voltage, emphasis on 'guess'
            # status_text = '%1.2fV' % (adc * 2.67793237653 / 0x0672)
            status_text = _hidpp20.BATTERY_STATUS.discharging

            if n.address == 0x00:
                status[_K.LIGHT_LEVEL] = None
                status.set_battery_info(charge, status_text, None)
            elif n.address == 0x10:
                status[_K.LIGHT_LEVEL] = lux
                if lux > 200:
                    status_text = _hidpp20.BATTERY_STATUS.recharging
                status.set_battery_info(charge, status_text, None)
            elif n.address == 0x20:
                if _log.isEnabledFor(_DEBUG):
                    _log.debug('%s: Light Check button pressed', device)
                status.changed(alert=_ALERT.SHOW_WINDOW)
                # first cancel any reporting
                # device.feature_request(_F.SOLAR_DASHBOARD)
                # trigger a new report chain
                reports_count = 15
                reports_period = 2  # seconds
                device.feature_request(_F.SOLAR_DASHBOARD, 0x00, reports_count, reports_period)
            else:
                _log.warn('%s: unknown SOLAR CHARGE %s', device, n)
        else:
            _log.warn('%s: SOLAR CHARGE not GOOD? %s', device, n)
        return True

    if feature == _F.TOUCHMOUSE_RAW_POINTS:
        if n.address == 0x00:
            if _log.isEnabledFor(_INFO):
                _log.info('%s: TOUCH MOUSE points %s', device, n)
        elif n.address == 0x10:
            touch = ord(n.data[:1])
            button_down = bool(touch & 0x02)
            mouse_lifted = bool(touch & 0x01)
            if _log.isEnabledFor(_INFO):
                _log.info('%s: TOUCH MOUSE status: button_down=%s mouse_lifted=%s', device, button_down, mouse_lifted)
        else:
            _log.warn('%s: unknown TOUCH MOUSE %s', device, n)
        return True

    if feature == _F.HIRES_WHEEL:
        if (n.address == 0x00):
            if _log.isEnabledFor(_INFO):
                flags, delta_v = _unpack('>bh', n.data[:3])
                high_res = (flags & 0x10) != 0
                periods = flags & 0x0f
                _log.info('%s: WHEEL: res: %d periods: %d delta V:%-3d', device, high_res, periods, delta_v)
            return True
        elif (n.address == 0x10):
            if _log.isEnabledFor(_INFO):
                flags = ord(n.data[:1])
                ratchet = flags & 0x01
                _log.info('%s: WHEEL: ratchet: %d', device, ratchet)
            return True
        else:
            _log.warn('%s: unknown WHEEL %s', device, n)
        return True

    _log.warn('%s: unrecognized %s for feature %s (index %02X)', device, n, feature, n.sub_id)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# USB ids of Logitech wireless receivers.
# Only receivers supporting the HID++ protocol can go in here.

from __future__ import absolute_import, division, print_function, unicode_literals

# max_devices is only used for receivers that do not support reading from _R.receiver_info offset 0x03, default to 1
# may_unpair is only used for receivers that do not support reading from _R.receiver_info offset 0x03, default to False
## should this last be changed so that may_unpair is used for all receivers? writing to _R.receiver_pairing doesn't seem right
# re_pairs determines whether a receiver pairs by replacing existing pairings, default to False
## currently only one receiver is so marked - should there be more?
# ex100_wpid_fix enable workarounds for EX100 and possible other old 27Mhz receivers

_DRIVER = ('hid-generic', 'generic-usb', 'logitech-djreceiver')

_unifying_receiver = lambda product_id: {
    'vendor_id': 0x046d,
    'product_id': product_id,
    'usb_interface': 2,
    'hid_driver': _DRIVER,  # noqa: F821
    'name': 'Unifying Receiver'
}

_nano_receiver = lambda product_id: {
    'vendor_id': 0x046d,
    'product_id': product_id,
    'usb_interface': 1,
    'hid_driver': _DRIVER,  # noqa: F821
    'name': 'Nano Receiver',
    'may_unpair': False,
    're_pairs': True
}

_nano_receiver_max2 = lambda product_id: {
    'vendor_id': 0x046d,
    'product_id': product_id,
    'usb_interface': 1,
    'hid_driver': _DRIVER,  # noqa: F821
    'name': 'Nano Receiver',
    'max_devices': 2,
    'may_unpair': False,
    're_pairs': True
}

_nano_receiver_maxn = lambda product_id, max: {
    'vendor_id': 0x046d,
    'product_id': product_id,
    'usb_interface': 1,
    'hid_driver': _DRIVER,  # noqa: F821
    'name': 'Nano Receiver',
    'max_devices': max,
    'may_unpair': False,
    're_pairs': True
}

_lenovo_receiver = lambda product_id: {
    'vendor_id': 0x17ef,
    'product_id': product_id,
    'usb_interface': 1,
    'hid_driver': _DRIVER,  # noqa: F821
    'name': 'Nano Receiver'
}

_lightspeed_receiver = lambda product_id: {
    'vendor_id': 0x046d,
    'product_id': product_id,
    'usb_interface': 2,
    'hid_driver': _DRIVER,  # noqa: F821
    'name': 'Lightspeed Receiver'
}

_ex100_receiver = lambda product_id: {
    'vendor_id': 0x046d,
    'product_id': product_id,
    'usb_interface': 1,
    'hid_driver': _DRIVER,  # noqa: F821
    'name': 'EX100 Receiver 27 Mhz',
    'max_devices': 4,
    'may_unpair': False,
    're_pairs': True,
    'ex100_wpid_fix': True
}

# standard Unifying receivers (marked with the orange Unifying logo)
UNIFYING_RECEIVER_C52B = _unifying_receiver(0xc52b)
UNIFYING_RECEIVER_C532 = _unifying_receiver(0xc532)

# Nano receviers that support the Unifying protocol
NANO_RECEIVER_ADVANCED = _nano_receiver(0xc52f)

# ex100 old style receiver pre-unifyimg protocol
EX100_27MHZ_RECEIVER_C517 = _ex100_receiver(0xc517)

# Nano receivers that don't support the Unifying protocol
NANO_RECEIVER_C518 = _nano_receiver(0xc518)
NANO_RECEIVER_C51A = _nano_receiver(0xc51a)
NANO_RECEIVER_C51B = _nano_receiver(0xc51b)
NANO_RECEIVER_C521 = _nano_receiver(0xc521)
NANO_RECEIVER_C525 = _nano_receiver(0xc525)
NANO_RECEIVER_C526 = _nano_receiver(0xc526)
NANO_RECEIVER_C52e = _nano_receiver(0xc52e)
NANO_RECEIVER_C531 = _nano_receiver(0xc531)
NANO_RECEIVER_C534 = _nano_receiver_max2(0xc534)
NANO_RECEIVER_C537 = _nano_receiver(0xc537)
NANO_RECEIVER_6042 = _lenovo_receiver(0x6042)

# Lightspeed receivers
LIGHTSPEED_RECEIVER_C539 = _lightspeed_receiver(0xc539)
LIGHTSPEED_RECEIVER_C53a = _lightspeed_receiver(0xc53a)
LIGHTSPEED_RECEIVER_C53f = _lightspeed_receiver(0xc53f)
LIGHTSPEED_RECEIVER_C53d = _lightspeed_receiver(0xc53d)

del _DRIVER, _unifying_receiver, _nano_receiver, _lenovo_receiver, _lightspeed_receiver

ALL = (
    UNIFYING_RECEIVER_C52B,
    UNIFYING_RECEIVER_C532,
    NANO_RECEIVER_ADVANCED,
    EX100_27MHZ_RECEIVER_C517,
    NANO_RECEIVER_C518,
    NANO_RECEIVER_C51A,
    NANO_RECEIVER_C51B,
    NANO_RECEIVER_C521,
    NANO_RECEIVER_C525,
    NANO_RECEIVER_C526,
    NANO_RECEIVER_C52e,
    NANO_RECEIVER_C531,
    NANO_RECEIVER_C534,
    NANO_RECEIVER_C537,
    NANO_RECEIVER_6042,
    LIGHTSPEED_RECEIVER_C539,
    LIGHTSPEED_RECEIVER_C53a,
    LIGHTSPEED_RECEIVER_C53f,
    LIGHTSPEED_RECEIVER_C53d,
)


def product_information(usb_id):
    if isinstance(usb_id, str):
        usb_id = int(usb_id, 16)
    for r in ALL:
        if usb_id == r.get('product_id'):
            return r
    return {}



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

import errno as _errno

from logging import INFO as _INFO
from logging import getLogger

from . import base as _base
from . import hidpp10 as _hidpp10
from . import hidpp20 as _hidpp20
from .base_usb import product_information as _product_information
from .common import strhex as _strhex
from .descriptors import DEVICES as _DESCRIPTORS
from .i18n import _
from .settings_templates import check_feature_settings as _check_feature_settings

_log = getLogger(__name__)
del getLogger

_R = _hidpp10.REGISTERS

#
#
#


class PairedDevice(object):
    def __init__(self, receiver, number, link_notification=None):
        assert receiver
        self.receiver = receiver

        assert number > 0 and number <= receiver.max_devices
        # Device number, 1..6 for unifying devices, 1 otherwise.
        self.number = number
        # 'device active' flag; requires manual management.
        self.online = None

        # the Wireless PID is unique per device model
        self.wpid = None
        self.descriptor = None

        # mouse, keyboard, etc (see _hidpp10.DEVICE_KIND)
        self._kind = None
        # Unifying peripherals report a codename.
        self._codename = None
        # the full name of the model
        self._name = None
        # HID++ protocol version, 1.0 or 2.0
        self._protocol = None
        # serial number (an 8-char hex string)
        self._serial = None

        self._firmware = None
        self._keys = None
        self._registers = None
        self._settings = None
        self._feature_settings_checked = False

        # Misc stuff that's irrelevant to any functionality, but may be
        # displayed in the UI and caching it here helps.
        self._polling_rate = None
        self._power_switch = None

        # if _log.isEnabledFor(_DEBUG):
        #     _log.debug("new PairedDevice(%s, %s, %s)", receiver, number, link_notification)

        if link_notification is not None:
            self.online = not bool(ord(link_notification.data[0:1]) & 0x40)
            self.wpid = _strhex(link_notification.data[2:3] + link_notification.data[1:2])
            # assert link_notification.address == (0x04 if unifying else 0x03)
            kind = ord(link_notification.data[0:1]) & 0x0F
            # fix EX100 wpid
            if receiver.ex100_wpid_fix:  # EX100 receiver
                self.wpid = _strhex(link_notification.data[2:3]) + '00'
                # workaround for EX100 switched kind
                if self.wpid == '3F00':
                    kind = 2
                if self.wpid == '6500':
                    kind = 1
            self._kind = _hidpp10.DEVICE_KIND[kind]
        else:
            # force a reading of the wpid
            pair_info = receiver.read_register(_R.receiver_info, 0x20 + number - 1)
            if pair_info:
                # may be either a Unifying receiver, or an Unifying-ready receiver
                self.wpid = _strhex(pair_info[3:5])
                kind = ord(pair_info[7:8]) & 0x0F
                self._kind = _hidpp10.DEVICE_KIND[kind]
            elif receiver.ex100_wpid_fix:
                # ex100 receiver, fill fake device_info with known wpid's
                # accordingly to drivers/hid/hid-logitech-dj.c
                # index 1 or 2 always mouse, index 3 always the keyboard,
                # index 4 is used for an optional separate numpad
                if number == 1:  # mouse
                    self.wpid = '3F00'
                    self._kind = _hidpp10.DEVICE_KIND[2]
                elif number == 3:  # keyboard
                    self.wpid = '6500'
                    self._kind = _hidpp10.DEVICE_KIND[1]
                else:  # unknown device number on EX100
                    _log.error('failed to set fake EX100 wpid for device %d of %s', number, receiver)
                    raise _base.NoSuchDevice(number=number, receiver=receiver, error='Unknown EX100 device')
            else:
                # unifying protocol not supported, must be a Nano receiver
                device_info = self.receiver.read_register(_R.receiver_info, 0x04)
                if device_info is None:
                    _log.error('failed to read Nano wpid for device %d of %s', number, receiver)
                    raise _base.NoSuchDevice(number=number, receiver=receiver, error='read Nano wpid')

                self.wpid = _strhex(device_info[3:5])
                self._power_switch = '(' + _('unknown') + ')'

        # the wpid is necessary to properly identify wireless link on/off notifications
        # also it gets set to None on this object when the device is unpaired
        assert self.wpid is not None, 'failed to read wpid: device %d of %s' % (number, receiver)

        self.descriptor = _DESCRIPTORS.get(self.wpid)
        if self.descriptor is None:
            # Last chance to correctly identify the device; many Nano receivers
            # do not support this call.
            codename = self.receiver.read_register(_R.receiver_info, 0x40 + self.number - 1)
            if codename:
                codename_length = ord(codename[1:2])
                codename = codename[2:2 + codename_length]
                self._codename = codename.decode('ascii')
                self.descriptor = _DESCRIPTORS.get(self._codename)

        if self.descriptor:
            self._name = self.descriptor.name
            self._protocol = self.descriptor.protocol
            if self._codename is None:
                self._codename = self.descriptor.codename
            if self._kind is None:
                self._kind = self.descriptor.kind

        if self._protocol is not None:
            self.features = None if self._protocol < 2.0 else _hidpp20.FeaturesArray(self)
        else:
            # may be a 2.0 device; if not, it will fix itself later
            self.features = _hidpp20.FeaturesArray(self)

    @property
    def protocol(self):
        if self._protocol is None and self.online is not False:
            self._protocol = _base.ping(self.receiver.handle, self.number)
            # if the ping failed, the peripheral is (almost) certainly offline
            self.online = self._protocol is not None

            # if _log.isEnabledFor(_DEBUG):
            #     _log.debug("device %d protocol %s", self.number, self._protocol)
        return self._protocol or 0

    @property
    def codename(self):
        if self._codename is None:
            codename = self.receiver.read_register(_R.receiver_info, 0x40 + self.number - 1)
            if codename:
                codename_length = ord(codename[1:2])
                codename = codename[2:2 + codename_length]
                self._codename = codename.decode('ascii')
                # if _log.isEnabledFor(_DEBUG):
                #     _log.debug("device %d codename %s", self.number, self._codename)
            else:
                self._codename = '? (%s)' % self.wpid
        return self._codename

    @property
    def name(self):
        if self._name is None:
            if self.online and self.protocol >= 2.0:
                self._name = _hidpp20.get_name(self)
        return self._name or self.codename or ('Unknown device %s' % self.wpid)

    @property
    def kind(self):
        if self._kind is None:
            pair_info = self.receiver.read_register(_R.receiver_info, 0x20 + self.number - 1)
            if pair_info:
                kind = ord(pair_info[7:8]) & 0x0F
                self._kind = _hidpp10.DEVICE_KIND[kind]
            elif self.online and self.protocol >= 2.0:
                self._kind = _hidpp20.get_kind(self)
        return self._kind or '?'

    @property
    def firmware(self):
        if self._firmware is None and self.online:
            if self.protocol >= 2.0:
                self._firmware = _hidpp20.get_firmware(self)
            else:
                self._firmware = _hidpp10.get_firmware(self)
        return self._firmware or ()

    @property
    def serial(self):
        if self._serial is None:
            serial = self.receiver.read_register(_R.receiver_info, 0x30 + self.number - 1)
            if serial:
                ps = ord(serial[9:10]) & 0x0F
                self._power_switch = _hidpp10.POWER_SWITCH_LOCATION[ps]
            else:
                # some Nano receivers?
                serial = self.receiver.read_register(0x2D5)

            if serial:
                self._serial = _strhex(serial[1:5])
            else:
                # fallback...
                self._serial = self.receiver.serial
        return self._serial or '?'

    @property
    def power_switch_location(self):
        if self._power_switch is None:
            ps = self.receiver.read_register(_R.receiver_info, 0x30 + self.number - 1)
            if ps is not None:
                ps = ord(ps[9:10]) & 0x0F
                self._power_switch = _hidpp10.POWER_SWITCH_LOCATION[ps]
            else:
                self._power_switch = '(unknown)'
        return self._power_switch

    @property
    def polling_rate(self):
        if self._polling_rate is None:
            pair_info = self.receiver.read_register(_R.receiver_info, 0x20 + self.number - 1)
            if pair_info:
                self._polling_rate = ord(pair_info[2:3])
            else:
                self._polling_rate = 0
        if self.online and self.protocol >= 2.0 and self.features and _hidpp20.FEATURE.REPORT_RATE in self.features:
            rate = _hidpp20.get_polling_rate(self)
            self._polling_rate = rate if rate is not None else self._polling_rate
        return self._polling_rate

    @property
    def keys(self):
        if self._keys is None:
            if self.online and self.protocol >= 2.0:
                self._keys = _hidpp20.get_keys(self) or ()
        return self._keys

    @property
    def registers(self):
        if self._registers is None:
            if self.descriptor and self.descriptor.registers:
                self._registers = list(self.descriptor.registers)
            else:
                self._registers = []
        return self._registers

    @property
    def settings(self):
        if self._settings is None:
            self._settings = []
            if self.descriptor and self.descriptor.settings:
                self._settings = []
                for s in self.descriptor.settings:
                    try:
                        setting = s(self)
                    except Exception as e:
                        # Do nothing if the device is offline
                        if self.online:
                            raise e
                    if setting is not None:
                        self._settings.append(setting)
        if not self._feature_settings_checked:
            self._feature_settings_checked = _check_feature_settings(self, self._settings)
        return self._settings

    def enable_notifications(self, enable=True):
        """Enable or disable device (dis)connection notifications on this
        receiver."""
        if not bool(self.receiver) or self.protocol >= 2.0:
            return False

        if enable:
            set_flag_bits = (
                _hidpp10.NOTIFICATION_FLAG.battery_status
                | _hidpp10.NOTIFICATION_FLAG.keyboard_illumination
                | _hidpp10.NOTIFICATION_FLAG.wireless
                | _hidpp10.NOTIFICATION_FLAG.software_present
            )
        else:
            set_flag_bits = 0
        ok = _hidpp10.set_notification_flags(self, set_flag_bits)
        if ok is None:
            _log.warn('%s: failed to %s device notifications', self, 'enable' if enable else 'disable')

        flag_bits = _hidpp10.get_notification_flags(self)
        flag_names = None if flag_bits is None else tuple(_hidpp10.NOTIFICATION_FLAG.flag_names(flag_bits))
        if _log.isEnabledFor(_INFO):
            _log.info('%s: device notifications %s %s', self, 'enabled' if enable else 'disabled', flag_names)
        return flag_bits if ok else None

    def request(self, request_id, *params, no_reply=False):
        return _base.request(self.receiver.handle, self.number, request_id, *params, no_reply=no_reply)

    read_register = _hidpp10.read_register
    write_register = _hidpp10.write_register

    def feature_request(self, feature, function=0x00, *params, no_reply=False):
        if self.protocol >= 2.0:
            return _hidpp20.feature_request(self, feature, function, *params, no_reply=no_reply)

    def ping(self):
        """Checks if the device is online, returns True of False"""
        protocol = _base.ping(self.receiver.handle, self.number)
        self.online = protocol is not None
        if protocol is not None:
            self._protocol = protocol
        return self.online

    def __index__(self):
        return self.number

    __int__ = __index__

    def __eq__(self, other):
        return other is not None and self.kind == other.kind and self.wpid == other.wpid

    def __ne__(self, other):
        return other is None or self.kind != other.kind or self.wpid != other.wpid

    def __hash__(self):
        return self.wpid.__hash__()

    __bool__ = __nonzero__ = lambda self: self.wpid is not None and self.number in self.receiver

    def __str__(self):
        return '<PairedDevice(%d,%s,%s,%s)>' % (self.number, self.wpid, self.codename or '?', self.serial)

    __unicode__ = __repr__ = __str__


#
#
#


class Receiver(object):
    """A Unifying Receiver instance.

    The paired devices are available through the sequence interface.
    """
    number = 0xFF
    kind = None

    def __init__(self, handle, device_info):
        assert handle
        self.handle = handle
        assert device_info
        self.path = device_info.path
        # USB product id, used for some Nano receivers
        self.product_id = device_info.product_id
        product_info = _product_information(self.product_id)
        if not product_info:
            raise Exception('Unknown receiver type', self.product_id)

        # read the serial immediately, so we can find out max_devices
        serial_reply = self.read_register(_R.receiver_info, 0x03)
        if serial_reply:
            self.serial = _strhex(serial_reply[1:5])
            self.max_devices = ord(serial_reply[6:7])
            # TODO _properly_ figure out which receivers do and which don't support unpairing
            # This code supposes that receivers that don't unpair support a pairing request for device index 0
            self.may_unpair = self.write_register(_R.receiver_pairing) is None
        else:  # handle receivers that don't have a serial number specially (i.e., c534)
            self.serial = None
            self.max_devices = product_info.get('max_devices', 1)
            self.may_unpair = product_info.get('may_unpair', False)

        self.name = product_info.get('name', '')
        self.re_pairs = product_info.get('re_pairs', False)
        self._str = '<%s(%s,%s%s)>' % (
            self.name.replace(' ', ''), self.path, '' if isinstance(self.handle, int) else 'T', self.handle
        )
        self.ex100_wpid_fix = product_info.get('ex100_wpid_fix', False)

        self._firmware = None
        self._devices = {}
        self._remaining_pairings = None

    def close(self):
        handle, self.handle = self.handle, None
        self._devices.clear()
        return (handle and _base.close(handle))

    def __del__(self):
        self.close()

    @property
    def firmware(self):
        if self._firmware is None and self.handle:
            self._firmware = _hidpp10.get_firmware(self)
        return self._firmware

    # how many pairings remain (None for unknown, -1 for unlimited)
    def remaining_pairings(self, cache=True):
        if self._remaining_pairings is None or not cache:
            ps = self.read_register(_R.receiver_connection)
            if ps is not None:
                ps = ord(ps[2:3])
                self._remaining_pairings = ps - 5 if ps >= 5 else -1
        return self._remaining_pairings

    def enable_notifications(self, enable=True):
        """Enable or disable device (dis)connection notifications on this
        receiver."""
        if not self.handle:
            return False

        if enable:
            set_flag_bits = (
                _hidpp10.NOTIFICATION_FLAG.battery_status
                | _hidpp10.NOTIFICATION_FLAG.wireless
                | _hidpp10.NOTIFICATION_FLAG.software_present
            )
        else:
            set_flag_bits = 0
        ok = _hidpp10.set_notification_flags(self, set_flag_bits)
        if ok is None:
            _log.warn('%s: failed to %s receiver notifications', self, 'enable' if enable else 'disable')
            return None

        flag_bits = _hidpp10.get_notification_flags(self)
        flag_names = None if flag_bits is None else tuple(_hidpp10.NOTIFICATION_FLAG.flag_names(flag_bits))
        if _log.isEnabledFor(_INFO):
            _log.info('%s: receiver notifications %s => %s', self, 'enabled' if enable else 'disabled', flag_names)
        return flag_bits

    def notify_devices(self):
        """Scan all devices."""
        if self.handle:
            if not self.write_register(_R.receiver_connection, 0x02):
                _log.warn('%s: failed to trigger device link notifications', self)

    def register_new_device(self, number, notification=None):
        if self._devices.get(number) is not None:
            raise IndexError('%s: device number %d already registered' % (self, number))

        assert notification is None or notification.devnumber == number
        assert notification is None or notification.sub_id == 0x41

        try:
            dev = PairedDevice(self, number, notification)
            assert dev.wpid
            if _log.isEnabledFor(_INFO):
                _log.info('%s: found new device %d (%s)', self, number, dev.wpid)
            self._devices[number] = dev
            return dev
        except _base.NoSuchDevice:
            _log.exception('register_new_device')

        _log.warning('%s: looked for device %d, not found', self, number)
        self._devices[number] = None

    def set_lock(self, lock_closed=True, device=0, timeout=0):
        if self.handle:
            action = 0x02 if lock_closed else 0x01
            reply = self.write_register(_R.receiver_pairing, action, device, timeout)
            if reply:
                return True
            _log.warn('%s: failed to %s the receiver lock', self, 'close' if lock_closed else 'open')

    def count(self):
        count = self.read_register(_R.receiver_connection)
        return 0 if count is None else ord(count[1:2])

    # def has_devices(self):
    #     return len(self) > 0 or self.count() > 0

    def request(self, request_id, *params):
        if bool(self):
            return _base.request(self.handle, 0xFF, request_id, *params)

    read_register = _hidpp10.read_register
    write_register = _hidpp10.write_register

    def __iter__(self):
        for number in range(1, 1 + self.max_devices):
            if number in self._devices:
                dev = self._devices[number]
            else:
                dev = self.__getitem__(number)
            if dev is not None:
                yield dev

    def __getitem__(self, key):
        if not bool(self):
            return None

        dev = self._devices.get(key)
        if dev is not None:
            return dev

        if not isinstance(key, int):
            raise TypeError('key must be an integer')
        if key < 1 or key > self.max_devices:
            raise IndexError(key)

        return self.register_new_device(key)

    def __delitem__(self, key):
        self._unpair_device(key, False)

    def _unpair_device(self, key, force=False):
        key = int(key)

        if self._devices.get(key) is None:
            raise IndexError(key)

        dev = self._devices[key]
        if not dev:
            if key in self._devices:
                del self._devices[key]
            return

        if self.re_pairs and not force:
            # invalidate the device, but these receivers don't unpair per se
            dev.online = False
            dev.wpid = None
            if key in self._devices:
                del self._devices[key]
            _log.warn('%s removed device %s', self, dev)
        else:
            reply = self.write_register(_R.receiver_pairing, 0x03, key)
            if reply:
                # invalidate the device
                dev.online = False
                dev.wpid = None
                if key in self._devices:
                    del self._devices[key]
                _log.warn('%s unpaired device %s', self, dev)
            else:
                _log.error('%s failed to unpair device %s', self, dev)
                raise IndexError(key)

    def __len__(self):
        return len([d for d in self._devices.values() if d is not None])

    def __contains__(self, dev):
        if isinstance(dev, int):
            return self._devices.get(dev) is not None

        return self.__contains__(dev.number)

    def __eq__(self, other):
        return other is not None and self.kind == other.kind and self.path == other.path

    def __ne__(self, other):
        return other is None or self.kind != other.kind or self.path != other.path

    def __hash__(self):
        return self.path.__hash__()

    def __str__(self):
        return self._str

    __unicode__ = __repr__ = __str__

    __bool__ = __nonzero__ = lambda self: self.handle is not None

    @classmethod
    def open(self, device_info):
        """Opens a Logitech Receiver found attached to the machine, by Linux device path.

        :returns: An open file handle for the found receiver, or ``None``.
        """
        try:
            handle = _base.open_path(device_info.path)
            if handle:
                return Receiver(handle, device_info)
        except OSError as e:
            _log.exception('open %s', device_info)
            if e.errno == _errno.EACCES:
                raise
        except Exception:
            _log.exception('open %s', device_info)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logging import getLogger  # , DEBUG as _DEBUG

from .common import FirmwareInfo as _FirmwareInfo
from .common import NamedInts as _NamedInts
from .common import bytes2int as _bytes2int
from .common import int2bytes as _int2bytes
from .common import strhex as _strhex
from .hidpp20 import BATTERY_STATUS, FIRMWARE_KIND

_log = getLogger(__name__)
del getLogger

#
# Constants - most of them as defined by the official Logitech HID++ 1.0
# documentation, some of them guessed.
#

DEVICE_KIND = _NamedInts(keyboard=0x01, mouse=0x02, numpad=0x03, presenter=0x04, trackball=0x08, touchpad=0x09)

POWER_SWITCH_LOCATION = _NamedInts(
    base=0x01,
    top_case=0x02,
    edge_of_top_right_corner=0x03,
    top_left_corner=0x05,
    bottom_left_corner=0x06,
    top_right_corner=0x07,
    bottom_right_corner=0x08,
    top_edge=0x09,
    right_edge=0x0A,
    left_edge=0x0B,
    bottom_edge=0x0C
)

# Some flags are used both by devices and receivers. The Logitech documentation
# mentions that the first and last (third) byte are used for devices while the
# second is used for the receiver. In practise, the second byte is also used for
# some device-specific notifications (keyboard illumination level). Do not
# simply set all notification bits if the software does not support it. For
# example, enabling keyboard_sleep_raw makes the Sleep key a no-operation unless
# the software is updated to handle that event.
# Observations:
# - wireless and software present were seen on receivers, reserved_r1b4 as well
# - the rest work only on devices as far as we can tell right now
# In the future would be useful to have separate enums for receiver and device notification flags,
# but right now we don't know enough.
# additional flags taken from https://drive.google.com/file/d/0BxbRzx7vEV7eNDBheWY0UHM5dEU/view?usp=sharing
NOTIFICATION_FLAG = _NamedInts(
    numpad_numerical_keys=0x800000,
    f_lock_status=0x400000,
    roller_H=0x200000,
    battery_status=0x100000,  # send battery charge notifications (0x07 or 0x0D)
    mouse_extra_buttons=0x080000,
    roller_V=0x040000,
    keyboard_sleep_raw=0x020000,  # system control keys such as Sleep
    keyboard_multimedia_raw=0x010000,  # consumer controls such as Mute and Calculator
    # reserved_r1b4=        0x001000,  # unknown, seen on a unifying receiver
    reserved5=0x008000,
    reserved4=0x004000,
    reserved3=0x002000,
    reserved2=0x001000,
    software_present=0x000800,  # .. no idea
    reserved1=0x000400,
    keyboard_illumination=0x000200,  # illumination brightness level changes (by pressing keys)
    wireless=0x000100,  # notify when the device wireless goes on/off-line
    mx_air_3d_gesture=0x000001,
)

ERROR = _NamedInts(
    invalid_SubID__command=0x01,
    invalid_address=0x02,
    invalid_value=0x03,
    connection_request_failed=0x04,
    too_many_devices=0x05,
    already_exists=0x06,
    busy=0x07,
    unknown_device=0x08,
    resource_error=0x09,
    request_unavailable=0x0A,
    unsupported_parameter_value=0x0B,
    wrong_pin_code=0x0C
)

PAIRING_ERRORS = _NamedInts(device_timeout=0x01, device_not_supported=0x02, too_many_devices=0x03, sequence_timeout=0x06)

BATTERY_APPOX = _NamedInts(empty=0, critical=5, low=20, good=50, full=90)
"""Known registers.
Devices usually have a (small) sub-set of these. Some registers are only
applicable to certain device kinds (e.g. smooth_scroll only applies to mice."""
REGISTERS = _NamedInts(
    # only apply to receivers
    receiver_connection=0x02,
    receiver_pairing=0xB2,
    devices_activity=0x2B3,
    receiver_info=0x2B5,

    # only apply to devices
    mouse_button_flags=0x01,
    keyboard_hand_detection=0x01,
    battery_status=0x07,
    keyboard_fn_swap=0x09,
    battery_charge=0x0D,
    keyboard_illumination=0x17,
    three_leds=0x51,
    mouse_dpi=0x63,

    # apply to both
    notifications=0x00,
    firmware=0xF1,
)
# Flags taken from https://drive.google.com/file/d/0BxbRzx7vEV7eNDBheWY0UHM5dEU/view?usp=sharing
DEVICE_FEATURES = _NamedInts(
    reserved1=0x010000,
    special_buttons=0x020000,
    enhanced_key_usage=0x040000,
    fast_fw_rev=0x080000,
    reserved2=0x100000,
    reserved3=0x200000,
    scroll_accel=0x400000,
    buttons_control_resolution=0x800000,
    inhibit_lock_key_sound=0x000001,
    reserved4=0x000002,
    mx_air_3d_engine=0x000004,
    host_control_leds=0x000008,
    reserved5=0x000010,
    reserved6=0x000020,
    reserved7=0x000040,
    reserved8=0x000080,
)

#
# functions
#


def read_register(device, register_number, *params):
    assert device, 'tried to read register %02X from invalid device %s' % (register_number, device)
    # support long registers by adding a 2 in front of the register number
    request_id = 0x8100 | (int(register_number) & 0x2FF)
    return device.request(request_id, *params)


def write_register(device, register_number, *value):
    assert device, 'tried to write register %02X to invalid device %s' % (register_number, device)
    # support long registers by adding a 2 in front of the register number
    request_id = 0x8000 | (int(register_number) & 0x2FF)
    return device.request(request_id, *value)


def get_battery(device):
    assert device
    assert device.kind is not None
    if not device.online:
        return
    """Reads a device's battery level, if provided by the HID++ 1.0 protocol."""
    if device.protocol and device.protocol >= 2.0:
        # let's just assume HID++ 2.0 devices do not provide the battery info in a register
        return

    for r in (REGISTERS.battery_status, REGISTERS.battery_charge):
        if r in device.registers:
            reply = read_register(device, r)
            if reply:
                return parse_battery_status(r, reply)
            return

    # the descriptor does not tell us which register this device has, try them both
    reply = read_register(device, REGISTERS.battery_charge)
    if reply:
        # remember this for the next time
        device.registers.append(REGISTERS.battery_charge)
        return parse_battery_status(REGISTERS.battery_charge, reply)

    reply = read_register(device, REGISTERS.battery_status)
    if reply:
        # remember this for the next time
        device.registers.append(REGISTERS.battery_status)
        return parse_battery_status(REGISTERS.battery_status, reply)


def parse_battery_status(register, reply):
    if register == REGISTERS.battery_charge:
        charge = ord(reply[:1])
        status_byte = ord(reply[2:3]) & 0xF0
        status_text = (
            BATTERY_STATUS.discharging if status_byte == 0x30 else
            BATTERY_STATUS.recharging if status_byte == 0x50 else BATTERY_STATUS.full if status_byte == 0x90 else None
        )
        return charge, status_text, None

    if register == REGISTERS.battery_status:
        status_byte = ord(reply[:1])
        charge = (
            BATTERY_APPOX.full if status_byte == 7  # full
            else BATTERY_APPOX.good if status_byte == 5  # good
            else BATTERY_APPOX.low if status_byte == 3  # low
            else BATTERY_APPOX.critical if status_byte == 1  # critical
            # pure 'charging' notifications may come without a status
            else BATTERY_APPOX.empty
        )

        charging_byte = ord(reply[1:2])
        if charging_byte == 0x00:
            status_text = BATTERY_STATUS.discharging
        elif charging_byte & 0x21 == 0x21:
            status_text = BATTERY_STATUS.recharging
        elif charging_byte & 0x22 == 0x22:
            status_text = BATTERY_STATUS.full
        else:
            _log.warn('could not parse 0x07 battery status: %02X (level %02X)', charging_byte, status_byte)
            status_text = None

        if charging_byte & 0x03 and status_byte == 0:
            # some 'charging' notifications may come with no battery level information
            charge = None

        # Return None for next charge level as this is not in HID++ 1.0 spec
        return charge, status_text, None


def get_firmware(device):
    assert device

    firmware = [None, None, None]

    reply = read_register(device, REGISTERS.firmware, 0x01)
    if not reply:
        # won't be able to read any of it now...
        return

    fw_version = _strhex(reply[1:3])
    fw_version = '%s.%s' % (fw_version[0:2], fw_version[2:4])
    reply = read_register(device, REGISTERS.firmware, 0x02)
    if reply:
        fw_version += '.B' + _strhex(reply[1:3])
    fw = _FirmwareInfo(FIRMWARE_KIND.Firmware, '', fw_version, None)
    firmware[0] = fw

    reply = read_register(device, REGISTERS.firmware, 0x04)
    if reply:
        bl_version = _strhex(reply[1:3])
        bl_version = '%s.%s' % (bl_version[0:2], bl_version[2:4])
        bl = _FirmwareInfo(FIRMWARE_KIND.Bootloader, '', bl_version, None)
        firmware[1] = bl

    reply = read_register(device, REGISTERS.firmware, 0x03)
    if reply:
        o_version = _strhex(reply[1:3])
        o_version = '%s.%s' % (o_version[0:2], o_version[2:4])
        o = _FirmwareInfo(FIRMWARE_KIND.Other, '', o_version, None)
        firmware[2] = o

    if any(firmware):
        return tuple(f for f in firmware if f)


def set_3leds(device, battery_level=None, charging=None, warning=None):
    assert device
    assert device.kind is not None
    if not device.online:
        return

    if REGISTERS.three_leds not in device.registers:
        return

    if battery_level is not None:
        if battery_level < BATTERY_APPOX.critical:
            # 1 orange, and force blink
            v1, v2 = 0x22, 0x00
            warning = True
        elif battery_level < BATTERY_APPOX.low:
            # 1 orange
            v1, v2 = 0x22, 0x00
        elif battery_level < BATTERY_APPOX.good:
            # 1 green
            v1, v2 = 0x20, 0x00
        elif battery_level < BATTERY_APPOX.full:
            # 2 greens
            v1, v2 = 0x20, 0x02
        else:
            # all 3 green
            v1, v2 = 0x20, 0x22
        if warning:
            # set the blinking flag for the leds already set
            v1 |= (v1 >> 1)
            v2 |= (v2 >> 1)
    elif charging:
        # blink all green
        v1, v2 = 0x30, 0x33
    elif warning:
        # 1 red
        v1, v2 = 0x02, 0x00
    else:
        # turn off all leds
        v1, v2 = 0x11, 0x11

    write_register(device, REGISTERS.three_leds, v1, v2)


def get_notification_flags(device):
    assert device

    # Avoid a call if the device is not online,
    # or the device does not support registers.
    if device.kind is not None:
        # peripherals with protocol >= 2.0 don't support registers
        if device.protocol and device.protocol >= 2.0:
            return

    flags = read_register(device, REGISTERS.notifications)
    if flags is not None:
        assert len(flags) == 3
        return _bytes2int(flags)


def set_notification_flags(device, *flag_bits):
    assert device

    # Avoid a call if the device is not online,
    # or the device does not support registers.
    if device.kind is not None:
        # peripherals with protocol >= 2.0 don't support registers
        if device.protocol and device.protocol >= 2.0:
            return

    flag_bits = sum(int(b) for b in flag_bits)
    assert flag_bits & 0x00FFFFFF == flag_bits
    result = write_register(device, REGISTERS.notifications, _int2bytes(flag_bits, 3))
    return result is not None


def get_device_features(device):
    assert device

    # Avoid a call if the device is not online,
    # or the device does not support registers.
    if device.kind is not None:
        # peripherals with protocol >= 2.0 don't support registers
        if device.protocol and device.protocol >= 2.0:
            return

    flags = read_register(device, REGISTERS.mouse_button_flags)
    if flags is not None:
        assert len(flags) == 3
        return _bytes2int(flags)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from threading import Timer as _Timer

from gi.repository import GLib, Gtk
from logitech_receiver.settings import KIND as _SETTING_KIND
from solaar.i18n import _
from solaar.ui import ui_async as _ui_async

#
#
#


def _read_async(setting, force_read, sbox, device_is_online):
    def _do_read(s, force, sb, online):
        v = s.read(not force)
        GLib.idle_add(_update_setting_item, sb, v, online, priority=99)

    _ui_async(_do_read, setting, force_read, sbox, device_is_online)


def _write_async(setting, value, sbox):
    _ignore, failed, spinner, control = sbox.get_children()
    control.set_sensitive(False)
    failed.set_visible(False)
    spinner.set_visible(True)
    spinner.start()

    def _do_write(s, v, sb):
        v = setting.write(v)
        GLib.idle_add(_update_setting_item, sb, v, True, priority=99)

    _ui_async(_do_write, setting, value, sbox)


def _write_async_key_value(setting, key, value, sbox):
    _ignore, failed, spinner, control = sbox.get_children()
    control.set_sensitive(False)
    failed.set_visible(False)
    spinner.set_visible(True)
    spinner.start()

    def _do_write_key_value(s, k, v, sb):
        v = setting.write_key_value(k, v)
        GLib.idle_add(_update_setting_item, sb, {k: v}, True, priority=99)

    _ui_async(_do_write_key_value, setting, key, value, sbox)


#
#
#


def _create_toggle_control(setting):
    def _switch_notify(switch, _ignore, s):
        if switch.get_sensitive():
            _write_async(s, switch.get_active() is True, switch.get_parent())

    c = Gtk.Switch()
    c.connect('notify::active', _switch_notify, setting)
    return c


def _create_choice_control(setting):
    def _combo_notify(cbbox, s):
        if cbbox.get_sensitive():
            _write_async(s, cbbox.get_active_id(), cbbox.get_parent())

    c = Gtk.ComboBoxText()
    # TODO i18n text entries
    for entry in setting.choices:
        c.append(str(int(entry)), str(entry))
    c.connect('changed', _combo_notify, setting)
    return c


def _create_map_choice_control(setting):
    def _map_value_notify_key(cbbox, s):
        setting, valueBox = s
        key_choice = int(cbbox.get_active_id())
        if cbbox.get_sensitive():
            valueBox.remove_all()
            _map_populate_value_box(valueBox, setting, key_choice)

    def _map_value_notify_value(cbbox, s):
        setting, keyBox = s
        key_choice = keyBox.get_active_id()
        if key_choice is not None and cbbox.get_sensitive() and cbbox.get_active_id():
            if setting._value.get(key_choice) != int(cbbox.get_active_id()):
                setting._value[key_choice] = int(cbbox.get_active_id())
                _write_async_key_value(setting, key_choice, setting._value[key_choice], cbbox.get_parent().get_parent())

    def _map_populate_value_box(valueBox, setting, key_choice):
        choices = None
        choices = setting.choices[key_choice]
        current = setting._value.get(str(key_choice)) if setting._value else None
        if choices:
            # TODO i18n text entries
            for choice in choices:
                valueBox.append(str(int(choice)), str(choice))
            if current is not None:
                valueBox.set_active_id(str(int(current)))

    c = Gtk.HBox(homogeneous=False, spacing=6)
    keyBox = Gtk.ComboBoxText()
    valueBox = Gtk.ComboBoxText()
    c.pack_start(keyBox, False, False, 0)
    c.pack_end(valueBox, False, False, 0)
    # TODO i18n text entries
    for entry in setting.choices:
        keyBox.append(str(int(entry)), str(entry))
    keyBox.set_active(0)
    keyBox.connect('changed', _map_value_notify_key, (setting, valueBox))
    _map_populate_value_box(valueBox, setting, int(keyBox.get_active_id()))
    valueBox.connect('changed', _map_value_notify_value, (setting, keyBox))
    return c


def _create_slider_control(setting):
    class SliderControl:
        __slots__ = ('gtk_range', 'timer', 'setting')

        def __init__(self, setting):
            self.setting = setting
            self.timer = None

            self.gtk_range = Gtk.Scale()
            self.gtk_range.set_range(*self.setting.range)
            self.gtk_range.set_round_digits(0)
            self.gtk_range.set_digits(0)
            self.gtk_range.set_increments(1, 5)
            self.gtk_range.connect('value-changed', lambda _, c: c._changed(), self)

        def _write(self):
            _write_async(self.setting, int(self.gtk_range.get_value()), self.gtk_range.get_parent())
            self.timer.cancel()

        def _changed(self):
            if self.gtk_range.get_sensitive():
                if self.timer:
                    self.timer.cancel()
                self.timer = _Timer(0.5, lambda: GLib.idle_add(self._write))
                self.timer.start()

    control = SliderControl(setting)
    return control.gtk_range


#
#
#


def _create_sbox(s):
    sbox = Gtk.HBox(homogeneous=False, spacing=6)
    sbox.pack_start(Gtk.Label(s.label), False, False, 0)

    spinner = Gtk.Spinner()
    spinner.set_tooltip_text(_('Working') + '...')

    failed = Gtk.Image.new_from_icon_name('dialog-warning', Gtk.IconSize.SMALL_TOOLBAR)
    failed.set_tooltip_text(_('Read/write operation failed.'))

    if s.kind == _SETTING_KIND.toggle:
        control = _create_toggle_control(s)
        sbox.pack_end(control, False, False, 0)
    elif s.kind == _SETTING_KIND.choice:
        control = _create_choice_control(s)
        sbox.pack_end(control, False, False, 0)
    elif s.kind == _SETTING_KIND.range:
        control = _create_slider_control(s)
        sbox.pack_end(control, True, True, 0)
    elif s.kind == _SETTING_KIND.map_choice:
        control = _create_map_choice_control(s)
        sbox.pack_end(control, True, True, 0)
    elif s.kind == _SETTING_KIND.multiple_toggle:
        # ugly temporary hack!
        choices = {k: [False, True] for k in s._validator.options}

        class X:
            def __init__(self, obj, ext):
                self.obj = obj
                self.ext = ext

            def __getattr__(self, attr):
                try:
                    return self.ext[attr]
                except KeyError:
                    return getattr(self.obj, attr)

        control = _create_map_choice_control(X(s, {'choices': choices}))
        sbox.pack_end(control, True, True, 0)
    else:
        raise Exception('NotImplemented')

    control.set_sensitive(False)  # the first read will enable it
    sbox.pack_end(spinner, False, False, 0)
    sbox.pack_end(failed, False, False, 0)

    if s.description:
        sbox.set_tooltip_text(s.description)

    sbox.show_all()
    spinner.start()  # the first read will stop it
    failed.set_visible(False)

    return sbox


def _update_setting_item(sbox, value, is_online=True):
    _ignore, failed, spinner, control = sbox.get_children()  # depends on box layout
    spinner.set_visible(False)
    spinner.stop()

    if value is None:
        control.set_sensitive(False)
        failed.set_visible(is_online)
        return

    control.set_sensitive(False)
    failed.set_visible(False)
    if isinstance(control, Gtk.Switch):
        control.set_active(value)
    elif isinstance(control, Gtk.ComboBoxText):
        control.set_active_id(str(int(value)))
    elif isinstance(control, Gtk.Scale):
        control.set_value(int(value))
    elif isinstance(control, Gtk.HBox):
        kbox, vbox = control.get_children()  # depends on box layout
        if value.get(kbox.get_active_id()):
            vbox.set_active_id(str(value.get(kbox.get_active_id())))
    else:
        raise Exception('NotImplemented')
    control.set_sensitive(True)


#
#
#

# config panel
_box = None
_items = {}


def create():
    global _box
    assert _box is None
    _box = Gtk.VBox(homogeneous=False, spacing=8)
    _box._last_device = None
    return _box


def update(device, is_online=None):
    assert _box is not None
    assert device
    device_id = (device.receiver.path, device.number)
    if is_online is None:
        is_online = bool(device.online)

    # if the device changed since last update, clear the box first
    if device_id != _box._last_device:
        _box.set_visible(False)
        _box._last_device = device_id

    # hide controls belonging to other devices
    for k, sbox in _items.items():
        sbox = _items[k]
        sbox.set_visible(k[0:2] == device_id)

    for s in device.settings:
        k = (device_id[0], device_id[1], s.name)
        if k in _items:
            sbox = _items[k]
        else:
            sbox = _items[k] = _create_sbox(s)
            _box.pack_start(sbox, False, False, 0)

        _read_async(s, False, sbox, is_online)

    _box.set_visible(True)


def clean(device):
    """Remove the controls for a given device serial.
    Needed after the device has been unpaired.
    """
    assert _box is not None
    device_id = (device.receiver.path, device.number)
    for k in list(_items.keys()):
        if k[0:2] == device_id:
            _box.remove(_items[k])
            del _items[k]


def destroy():
    global _box
    _box = None
    _items.clear()
